from timeit import default_timer as timer

import numpy as np
import pathlib
import typing as t
from abc import ABC, abstractmethod


class Producer(ABC):
    """
    This is responsible for creating segmentation that will be later evaluated.
    """
    TAG_PREFIX = 'producer_'
    RUN_PREFIX = 'run_'

    def __init__(self, name):
        self.name = name

    @abstractmethod
    def segmentation(self, input_data: t.Union[np.ndarray, pathlib.Path], input_tag: dict) \
            -> t.Union[np.ndarray, pathlib.Path]:
        """
        Args:
            input_data: path to input data or image in uint8 (0-255)
            input_tag: tags of the input data
        """
        pass

    def calculate(self, input_data: t.Union[np.ndarray, pathlib.Path], input_tag: dict) \
            -> (t.Union[np.ndarray, pathlib.Path], dict):
        start_time = timer()
        seg = self.segmentation(input_data, input_tag)

        seg_tag = {}
        prediction_time = timer() - start_time
        seg_tag[f'{Producer.RUN_PREFIX}time'] = prediction_time
        seg_tag[f'{Producer.RUN_PREFIX}fps'] = round(1.0 / prediction_time, 2)
        seg_tag[f'{Producer.TAG_PREFIX}name'] = self.name
        seg_tag[f'{Producer.TAG_PREFIX}details'] = self.__repr__()
        return seg, seg_tag

    def __repr__(self):
        return f"{self.__class__} ({self.__dict__})"

    def __str__(self):
        return f"{self.name}"



import os
import unittest
from pathlib import Path
import numpy as np
import numpy.testing as nptest

import sep._commons.imgutil as imgutil
from sep.loaders.images import ImagesLoader, SepCannotDetermineAnnotationPathError
from tests.testbase import TestBase


class TestImagesLoader(TestBase):
    def test_loading(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics/lights"))
        self.assertEqual(4, len(loader))
        self.assertEqual(['lights01', 'lights02', 'lights03', 'lights04'], loader.input_order)

        input_data_02_by_id = loader.load_image(1)
        input_data_02_by_name = loader.load_image('lights02')
        nptest.assert_equal(input_data_02_by_id, input_data_02_by_name)

        tag_02 = loader.load_tag('lights02')
        self.assertEqual("lights02", tag_02["id"])
        self.assertEqual("thenet", tag_02["source"])
        non_existing_tag10 = loader.load_tag('lights10')
        self.assertEqual("lights10", non_existing_tag10["id"])
        self.assertNotIn("source", non_existing_tag10)

        annotation_1 = loader.load_annotation(0)
        self.assertEqual(annotation_1.shape, input_data_02_by_id.shape[:2])
        self.assertEqual(255, annotation_1.max())

        tag_1 = loader.load_tag(0)
        self.assertEqual(0, tag_1["id"])  # TODO RETHINK default tags mirror exact call

    def test_get_element(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics/lights"))
        second_elem = loader[1]
        self.assertIn("image", second_elem)
        self.assertIn("annotation", second_elem)
        self.assertIn("tag", second_elem)

    def test_iterate_through(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics/lights"))
        data = [p for p in loader]
        self.assertEqual(4, len(data))
        second_elem = data[1]
        self.assertIn("image", second_elem)
        self.assertIn("annotation", second_elem)
        self.assertIn("tag", second_elem)

    def test_relative(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics"))
        data_names = loader.list_images()
        self.assertEqual(7, len(data_names))
        self.assertEqual("human_1", data_names[0])
        self.assertEqual(os.path.join("humans", "human_1.tif"), loader.get_relative_path(0))
        self.assertEqual(os.path.join("humans", "human_1.tif"), loader.get_relative_path("human_1"))

    def test_listing_save(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics"))
        data_names = loader.list_images()
        self.assertEqual(7, len(data_names))

        listing_path = self.add_temp("loader_listing.txt")
        loader.save(listing_path, add_tag_path=False)

        # check that there are 5 lines and that they point to the actual files
        with open(listing_path, "r") as listing_file:
            listing_lines = listing_file.readlines()
        self.assertEqual(7, len(listing_lines))
        self.assertEqual(f"{Path('humans/human_1.tif')}, {Path('humans/human_1_gt.png')}",
                         listing_lines[0].strip())

        loader.save(listing_path, add_tag_path=True)
        # now it has tag files (at least expected)
        with open(listing_path, "r") as listing_file:
            listing_lines = listing_file.readlines()
        self.assertEqual(f"{Path('humans/human_1.tif')}, {Path('humans/human_1_gt.png')}, {Path('humans/human_1.json')}",
                         listing_lines[0].strip())

    def test_listing_filter_load(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics"))
        self.assertEqual(7, len(loader.list_images()))
        names = loader.list_images()
        loader.show_summary()
        loader.filter_files([0, 'human_3', 4])
        with self.assertRaises(ValueError):
            loader.filter_files([0, 'my_image'])
        self.assertEqual(3, len(loader.list_images()))
        self.assertEqual(names[0], loader.list_images()[0])
        self.assertEqual(names[2], loader.list_images()[1])
        self.assertEqual(names[4], loader.list_images()[2])

        listing_file = self.add_temp("loader_listing.txt")
        loader.save(listing_file, add_tag_path=False)

        # check that there are 3 lines and that they point to the actual files
        loader_reloaded = ImagesLoader.from_listing(self.root_test_dir("input/basics"), filepath=listing_file)
        self.assertEqual(3, len(loader_reloaded.list_images()))
        self.assertEqual(loader.list_images(), loader_reloaded.list_images())

        # check filter using other int types
        loader.filter_files([0, 1])
        loader.filter_files(np.array([0, 1], dtype=int))
        loader.filter_files(np.array([0, 1], dtype=np.int64))
        loader.filter_files(np.array([0, 1], dtype=np.int32))
        loader.filter_files(np.array([0, 1], dtype=np.int16))
        loader.show_summary()

    def test_listing_load_partial_gt(self):
        with self.create_temp("loader_listing.txt") as listing_file:
            listing_file.writelines(f"{Path('humans/human_1.tif')}, {Path('humans/human_1_gt.png')}\n")
            listing_file.writelines(f"{Path('humans/human_2.tif')}, \n")
            # TODO there should be difference between not known ground truth path and non existing file at known position
            # TODO at the moment there is none - there is no trace of what was in the listing file

        loader_proper = ImagesLoader.from_listing(self.root_test_dir("input/basics"), filepath="loader_listing.txt")
        self.assertEqual(2, len(loader_proper))
        elem_1 = loader_proper[0]
        self.assertIsNotNone(elem_1['image'])
        self.assertIsNotNone(elem_1['annotation'])
        elem_2 = loader_proper[1]
        self.assertIsNotNone(elem_2['image'])
        self.assertIsNone(elem_2['annotation'])

    def test_listing_load_missing_image(self):
        with self.create_temp("loader_listing.txt") as listing_file:
            listing_file.writelines(f"{Path('humans/human_1.tif')}, {Path('humans/human_1_gt.png')}\n")
            listing_file.writelines(f"{Path('humans2/human2_2.tif')}, {Path('humans2/human2_2_gt.png')}\n")

        # always fail for missing image TODO TMP think should work but splits have to be handle somehow
        # with self.assertRaises(ValueError):
        #    ImagesLoader.from_listing(self.root_test_dir("input"), filepath="loader_listing.txt")

    def test_listing_load_missing_annotation_tag(self):
        with self.create_temp("loader_listing.txt") as listing_file:
            listing_file.writelines(f"{Path('lights/lights01.tif')},\n")
            listing_file.writelines(f"{Path('lights/lights02.tif')}\n")
            listing_file.writelines(f"{Path('humans/human_1.tif')}, {Path('humans/human_1_gt.png')}\n")
            listing_file.writelines(
                f"{Path('humans/human_2.tif')}, {Path('humans/human_2_gt_extra.png')}, {Path('humans/human_2_gt_extra.json')}\n")

        # this should work - TODO potentially log warnings
        loader = ImagesLoader.from_listing(self.root_test_dir("input/basics"), filepath="loader_listing.txt")
        self.assertEqual(4, len(loader.list_images()))
        self.assertIsNone(loader['lights01']['annotation'])
        self.assertIsNone(loader['lights02']['annotation'])
        self.assertIsNotNone(loader['human_1']['annotation'])
        self.assertIsNone(loader['human_2']['annotation'])
        self.assertNotIn("source", loader['human_2']['tag'])

    def test_listing_load_from_actual_file(self):
        loader = ImagesLoader.from_listing(self.root_test_dir("input/basics"),
                                           filepath=self.root_test_dir("input/picky_images.txt"))
        images_names = loader.list_images()
        expected_names = ['human_1', 'human_3', 'lights01', 'lights02']
        self.assertEqual(expected_names, images_names)

        data = list(loader)
        self.assertEqual(4, len(data))
        for d in data[:3]:
            self.assertIsNotNone(d['image'])
            self.assertGreater(d['image'].shape[0], 0)
            self.assertIsNotNone(d['annotation'])
            self.assertGreater(d['annotation'].shape[0], 0)

        # first one has also tag json file
        self.assertIn("source", data[0]['tag'])
        self.assertIsNotNone(data[0]['tag'])

        # second does not have json file
        self.assertNotIn("source", data[1]['tag'])
        self.assertIsNotNone(data[1]['tag'])

        # last one does not have annotation specified
        self.assertIsNotNone(data[3]['image'])
        self.assertGreater(data[3]['image'].shape[0], 0)
        self.assertIsNone(data[3]['annotation'])
        self.assertNotIn("source", data[3]['tag'])
        self.assertIsNotNone(data[3]['tag'])

    def test_extended_ids(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics"))
        first_image = loader.list_images()[0]
        self.assertEqual("human_1", first_image)
        first_path = loader.list_images_paths()[0]
        self.assertEqual("human_1", loader.path_to_id(first_path))

        loader.extended_ids = True
        self.assertEqual("humans/human_1", loader.path_to_id(first_path))

    def test_use_extended_ids_when_necessary(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/brightfield"), annotation_extension=".tif")
        image_names = loader.list_images()
        self.assertEqual(2, len(image_names))
        self.assertEqual("view1/BF_frame001", image_names[0])
        self.assertEqual("view2/BF_frame001", image_names[1])
        # make sure that gt is read despite being tif
        self.assertIsNotNone(loader.load_annotation(0))
        self.assertIsNotNone(loader.load_annotation(1))

    def test_custom_annotation_mapping(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/brightfield/view1"),
                                        annotation_suffix=None,
                                        annotation_for_image_finder=lambda p: p.stem + "_gt.tif")
        # Find annotation to the input image but assumed that annotation is a sample on her own.
        image_names = loader.list_images()
        self.assertEqual(2, len(image_names))
        self.assertEqual("BF_frame001", image_names[0])
        self.assertEqual("BF_frame001_gt", image_names[1])
        self.assertIsNotNone(loader.load_annotation(0))
        self.assertIsNone(loader.load_annotation(1))

        loader = ImagesLoader.from_tree(self.root_test_dir("input/brightfield/view1"),
                                        annotation_suffix=None,
                                        annotation_checker=lambda p: p.stem.endswith("_gt"),
                                        annotation_for_image_finder=lambda p: p.stem + "_gt.tif")
        # Now finds proper dataset.
        image_names = loader.list_images()
        self.assertEqual(1, len(image_names))
        self.assertEqual("BF_frame001", image_names[0])
        self.assertIsNotNone(loader.load_annotation(0))

    def test_save_tag_no_tag_file(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/brightfield"), annotation_extension=".tif")
        sample_name = loader.list_images()[0]
        sample_image_path = loader.input_paths[sample_name]
        sample_tag_path = loader.json_tags.get(sample_name)
        self.assertEqual(sample_tag_path, None)
        new_tag = {"review": "Rejected"}
        new_tag['id'] = loader.load_tag(sample_name)['id']
        loader.save_tag(sample_name, new_tag)

        new_tag_path = loader.json_tags.get(sample_name)
        self.add_temp(new_tag_path)
        self.assertEqual(new_tag_path, Path(sample_image_path).with_suffix(".json"))

        # Now reload with new loader
        loader2 = ImagesLoader.from_tree(self.root_test_dir("input/brightfield"), annotation_extension=".tif")
        self.assertEqual(new_tag_path, loader2.json_tags.get(sample_name))
        self.assertDictEqual(new_tag, loader2.load_tag(sample_name))

    def test_save_tag_existing_tag_file(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input/basics/humans"))
        sample_name = loader.list_images()[0]
        self.assertEqual("human_1", sample_name)
        sample_tag_path = loader.json_tags.get(sample_name)
        self.schedule_restoration(sample_tag_path)
        self.assertIsNotNone(sample_tag_path)
        new_tag = {"review": "Rejected"}
        loader.extend_tag(sample_name, new_tag)

        # Now reload with new loader
        loader2 = ImagesLoader.from_tree(self.root_test_dir("input/basics/humans"), annotation_extension=".tif")
        self.assertSubset(loader2.load_tag(sample_name), new_tag)

    def test_save_annotation_no_path(self):
        # Based on tree mode
        loader = ImagesLoader.from_tree(self.root_test_dir("input/giraffes"), input_extensions=".jpg")
        self.assertEqual(3, len(loader))
        sample_name = "giraffes_001"
        giraffes_001_sample = loader[sample_name]

        new_annotation = self.random_uint(imgutil.get_2d_size(giraffes_001_sample['image']))
        self.assertIsNone(loader.load_annotation(sample_name))
        loader.save_annotation(sample_name, new_annotation)
        self.np_assert_equal(new_annotation, loader.load_annotation(sample_name))  # reload and compare
        self.add_temp(loader.annotation_paths[sample_name])

        # Now reload with new loader
        loader2 = ImagesLoader.from_tree(self.root_test_dir("input/giraffes"), input_extensions=".jpg")
        nptest.assert_equal(new_annotation, loader2.load_annotation(sample_name))

        # Now try to use same in listing and fail.
        loader_listing = ImagesLoader.from_listing(self.root_test_dir("input/basics"),
                                                   filepath=self.root_test_dir("input/picky_images.txt"))
        sample_name = "lights02"
        lights02_sample = loader_listing[sample_name]
        new_annotation = self.random_uint(imgutil.get_2d_size(lights02_sample['image']))
        self.assertIsNone(loader_listing.load_annotation(sample_name))
        with self.assertRaises(SepCannotDetermineAnnotationPathError):
            loader_listing.save_annotation(sample_name, new_annotation)

        # Now provide explicit new path for annotations.
        # TODO

    def test_save_annotation_existing_path(self):
        loader = ImagesLoader.from_listing(self.root_test_dir("input/basics"),
                                           filepath=self.root_test_dir("input/picky_images.txt"))
        sample_name = "lights01"
        lights01_sample = loader[sample_name]
        new_annotation = self.random_uint(imgutil.get_2d_size(lights01_sample['image']))
        old_annotation = lights01_sample['annotation']
        self.assertIsNotNone(old_annotation)
        self.np_assert_not_equal(new_annotation, old_annotation)
        self.schedule_restoration(loader.annotation_paths[sample_name])

        loader.save_annotation(sample_name, new_annotation)
        nptest.assert_equal(new_annotation, loader.load_annotation(sample_name))

        # Now reload with new loader
        loader2 = ImagesLoader.from_listing(self.root_test_dir("input/basics"),
                                            filepath=self.root_test_dir("input/picky_images.txt"))
        nptest.assert_equal(new_annotation, loader2.load_annotation(sample_name))

    def test_dot_in_names_and_folders(self):
        loader = ImagesLoader.from_tree(self.root_test_dir("input"))
        loader.show_summary()

        names_with_ptak = [n for n in loader.list_images() if 'pta' in n]
        paths_with_ptak = [n for n in loader.list_images_paths() if 'pta' in str(n)]
        self.assertEqual(3, len(names_with_ptak))
        print(names_with_ptak)
        print(paths_with_ptak)
        expected_names = ['ptak.sep/kasia.imi_06148fb', 'ptak.sep/kasia.imi_06157fb', 'ptak.sep/sep-plowy']
        for n in names_with_ptak:
            sample = loader[n]
            self.assertIn(n, expected_names)
            self.assertIsNotNone(sample['image'])
            self.assertTrue(sample['image'].mean() > 0.1)
            self.assertIsNotNone(sample['annotation'])
            self.assertTrue(sample['annotation'].mean() > 0.1)


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from enum import Enum

import napari
import numpy as np
from magicgui import magicgui

from sep._commons.imgutil import get_2d_size


class Inspector:
    # Base class which wraps the common extension paths yet still allow to use full power of napari.
    # TODO docs
    # TODO add standard tags usage
    # TODO add information about the name of the current file
    def __init__(self, samples_collection, additional_tags: dict = None, loop_images=True, hide_left=True):
        self.samples_collection = samples_collection
        self.collection_size = len(samples_collection)
        self.viewer_state = {"current": 0, 'current_tag': {}, 'temporary_tags': {}}
        self.current_tags_state = self.viewer_state['temporary_tags']
        self.additional_tags = additional_tags or dict()
        self.load_tag_to_control = None
        self.viewer = None
        self.extra_labels = {}
        self.sample_changed = None
        self.loop_images = loop_images
        self.hide_left = hide_left

    def set_load_tag_to_control(self, load_tag_to_control):
        self.load_tag_to_control = load_tag_to_control

    def get_or_copy_tags(self, sample_num):
        if sample_num not in self.current_tags_state:
            self.current_tags_state[sample_num] = deepcopy(self.samples_collection[sample_num]['tag'])
            if self.additional_tags is not None:  # add new tags only for inspection?
                self.current_tags_state[sample_num].update(self.additional_tags)
        return self.current_tags_state[sample_num]

    def get_labels_or_empty(self, sample_num):
        annotation = self.samples_collection[sample_num]['annotation']
        if annotation is None:  # fill in with dummy
            image = self.samples_collection[sample_num]['image']
            annotation = np.zeros(get_2d_size(image), dtype=np.uint8)
        return annotation

    @property
    def viewer_created(self):
        return self.viewer is not None

    def add_extra_label(self, name, function):
        self.extra_labels[name] = function
        self.extra_images[name] = self.viewer.add_labels(function(self.get_labels_or_empty(0)),
                                                         name=name, opacity=0.8)

    def create_viewer(self, dock_area_sample="bottom", dock_area_refresh="bottom"):
        self.viewer = napari.Viewer()
        self.input_image = self.viewer.add_image(self.samples_collection[0]['image'], rgb=True)
        self.label_image = self.viewer.add_labels(self.get_labels_or_empty(0))
        self.extra_images = {}

        @magicgui(
            auto_call=True,
            Sample={"widget_type": "IntSlider", "max": self.collection_size - 1, "readout": True}
        )
        def change_sample(Sample: int):
            self.set_sample(Sample)

        # Stop it from getting focus after each sample change.
        change_sample.Sample.changed.connect(lambda val: change_sample.Sample.native.setFocus())

        @magicgui(call_button='Refresh')
        def refresh():
            self.set_sample(self.viewer_state["current"])

        self.set_sample(0)
        self.viewer.window.add_dock_widget(change_sample, area=dock_area_sample)
        self.viewer.window.add_dock_widget(refresh, area=dock_area_refresh)

        if self.hide_left:
            self.viewer.window._qt_viewer.dockLayerList.setVisible(False)
            self.viewer.window._qt_viewer.dockLayerControls.setVisible(False)

        @self.viewer.bind_key('z')
        def prev_label(event=None):
            if loop_images:
                change_sample.Sample.value = (self.viewer_state["current"] - 1 + self.collection_size) % self.collection_size
            else:
                change_sample.Sample.value = max(0, self.viewer_state["current"] - 1)

        @self.viewer.bind_key('c')
        def next_label(event=None):
            if loop_images:
                change_sample.Sample.value = (self.viewer_state["current"] + 1 + self.collection_size) % self.collection_size
            else:
                change_sample.Sample.value = min(self.collection_size - 1, self.viewer_state["current"] + 1)

    def set_sample(self, new_cur):
        current = (new_cur + self.collection_size) % self.collection_size
        self.viewer_state["current"] = current

        current_im = self.samples_collection[current]['image']
        current_ann = self.get_labels_or_empty(current)
        self.input_image.data, self.label_image.data = current_im, current_ann
        for name in self.extra_images:
            if self.extra_images[name].visible:
                self.extra_images[name].data = self.extra_labels[name](current_ann)
            else:
                self.extra_images[name].data[:] = 0

        self.viewer_state['current_tag'] = self.get_or_copy_tags(current)
        if self.load_tag_to_control:
            self.load_tag_to_control(self.viewer_state['current_tag'])

        if self.sample_changed:
            self.sample_changed(current, current_im, current_ann)


class ReviewEnum(Enum):
    No = 0
    Approved = 1
    Rejected = 2


def add_review_option(inspector):
    inspector.additional_tags['review'] = ReviewEnum.No

    # Extend inspector with the additional tag representing the sample review state.
    def load_tags(current_tag):
        tags_properties.Review.value = inspector.viewer_state['current_tag'].get('review', ReviewEnum.No)

    @magicgui(auto_call=True)
    def tags_properties(Review: ReviewEnum):
        inspector.viewer_state['current_tag']['review'] = Review

    inspector.set_load_tag_to_control(load_tags)
    assert inspector.viewer_created, "Viewer has to be crated for binding new keys."

    @inspector.viewer.bind_key('1')
    def approve_move_key(event=None):
        tags_properties.Review.value = ReviewEnum.Approved
        inspector.viewer.keymap['C']()  # move to the next

    @inspector.viewer.bind_key('2')
    def reject_move_key(event=None):
        tags_properties.Review.value = ReviewEnum.Rejected
        inspector.viewer.keymap['C']()  # move to the next

    @inspector.viewer.bind_key('`')
    def back_move_key(event=None):
        inspector.viewer.keymap['Z']()  # move to the next

    inspector.viewer.window.add_dock_widget(tags_properties)

    return inspector



import numbers
import traceback

import math
import numpy as np
import typing as t
from abc import ABC, abstractmethod

from sep._commons.movies import StreamReader
from sep._commons.utils import *
from sep.loaders.files import FilesLoader
from sep.loaders.loader import Loader


class FrameSelector(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def select(self, samples_frames: t.List[int], framerate: float = None) -> t.Iterable[t.Tuple[t.List[int], int]]:
        """
        Select subset of the provided frame positions.
        Args:
            framerate: optional video framerate in frames per sec
            samples_frames: list of frame positions of the samples

        Returns:
            The set of clips where each clip is (subset list, clip_id).
        """
        pass


class FrameByGroupSelector(FrameSelector):
    def __init__(self, clips_len=99999999, clips_skip=0, skip_first=0, skip_last=0):
        super().__init__()
        self.skip_last = skip_last
        self.skip_first = skip_first
        self.clips_skip = clips_skip
        self.clips_len = clips_len

    def select(self, samples_frames: t.List[int], framerate: float = None) -> t.Iterable[t.Tuple[t.List[int], int]]:
        samples = list(samples_frames[self.skip_first:len(samples_frames) - self.skip_last])
        clip_nr = 0
        while samples:
            yield samples[:self.clips_len], clip_nr
            del samples[:self.clips_len + self.clips_skip]
            clip_nr += 1


class FrameByIntervalSelector(FrameSelector):
    def __init__(self, frames_to_extract: int, interval_start_time: float = 0, interval_end_time: float = 99999):
        super().__init__()
        self.interval_start = interval_start_time
        self.interval_end = interval_end_time
        self.frames_to_extract = frames_to_extract

    def select(self, samples_frames: t.List[int], framerate: float = 1) -> t.Iterable[t.Tuple[t.List[int], int]]:
        start_pos = math.floor(self.interval_start * framerate)
        end_pos = math.ceil(self.interval_end * framerate)
        inside_frames = [s for s in samples_frames if start_pos <= s <= end_pos]

        # Split it into frames_to_extract + 1 spaces.
        step_float = len(inside_frames) / (self.frames_to_extract + 1)
        current = step_float

        clip_nr = 0
        while current < len(inside_frames) and clip_nr < self.frames_to_extract:
            current_pos = int(current)
            yield [inside_frames[current_pos]], clip_nr
            clip_nr += 1
            current += step_float


class MoviesLoader(Loader):
    """
    This one loads frames from the movies files and tags them so that they can be reorganizes into the files
    after evaluation. It also provides timestamp so that real time solution can be tested as well.

    For this the number of the image is the number in generated set of frames.
    The name is for example dragoons_00013.mp4 (frame 13 of the movie clip dragoons).

    This can be fairly slow if you have random access to images.

    Beware it is blocking related videos until it is closed.
    """
    MOVIE_TAG_PREFIX = 'movie_'
    USER_TAG_PREFIX = 'user_'

    def __init__(self, data_root, framerate, clips_len=99999999, clips_skip=0, annotation_as_mask=False, verbose=0):
        super().__init__()
        self.annotation_as_mask = annotation_as_mask
        self.data_root = data_root
        self.verbose = verbose
        self.selector = FrameByGroupSelector(clips_len, clips_skip)
        self.framerate = framerate
        self.video_image_reader: t.Optional[StreamReader] = None
        self.video_annotation_reader: t.Optional[StreamReader] = None
        self.input_paths = {}
        self.annotation_paths = {}
        self.json_tags = {}
        self.files_loader = None

    @classmethod
    def from_tree(cls, data_root,
                  framerate, clips_len=99999999, clips_skip=0, annotation_as_mask=False, verbose=0,
                  input_extensions=None,
                  annotation_suffix='_gt', annotation_extension='.mp4',
                  annotation_for_movie_finder: t.Callable[[pathlib.Path], str] = None):
        self = cls(data_root=data_root, framerate=framerate, clips_len=clips_len, clips_skip=clips_skip,
                   annotation_as_mask=annotation_as_mask, verbose=verbose)
        input_extensions = input_extensions or ['.mov', '.mp4', '.mpg', '.avi']
        files_loader = FilesLoader.from_tree(data_root, input_extensions=input_extensions,
                                             annotation_extension=annotation_extension,
                                             annotation_suffix=annotation_suffix,
                                             annotation_for_image_finder=annotation_for_movie_finder,
                                             verbose=verbose)
        self.generate_frames(files_loader)
        return self

    @classmethod
    def from_listing(cls, data_root, filepath,
                     framerate, clips_len=99999999, clips_skip=0, annotation_as_mask=False, verbose=0,
                     validate_list=False):
        self = cls(data_root=data_root, framerate=framerate, clips_len=clips_len, clips_skip=clips_skip,
                   annotation_as_mask=annotation_as_mask, verbose=verbose)
        files_loader = FilesLoader.from_listing(data_root, filepath, verbose, validate_list)
        self.generate_frames(files_loader)
        return self

    def show_summary(self):
        print(self)
        print(f"Found {len(self.input_paths)} input frames.")
        print(f"Found {len(self.annotation_paths)} annotation frames.")
        print(f"Found {len(self.json_tags)} tags.")

    def generate_frames(self, files_loader: FilesLoader):
        self.input_paths = {}
        self.annotation_paths = {}
        self.json_tags = {}
        self.files_loader = files_loader

        for movie_path in self.list_movies_paths():
            movie_id = self.files_loader.path_to_id(movie_path)
            annotation_movie_path = self.files_loader.load_annotation(movie_id)
            movie_tag = self.files_loader.load_tag(movie_id)

            # Process input data movies.
            movie_frames = MoviesLoader.load_movie_images(movie_path, self.framerate, self.selector)
            for frame_id, tag in zip(movie_frames['images'], movie_frames['tags']):
                frame_path = f"{movie_path}{frame_id}"
                frame_id = f"{movie_id}{frame_id}"
                tag['id'] = frame_id
                update_with_suffix(tag, movie_tag, prefix=MoviesLoader.MOVIE_TAG_PREFIX)

                self.input_paths[frame_id] = frame_path
                self.json_tags[frame_id] = tag

            # Process annotation movies.
            if annotation_movie_path:
                annotation_frames = MoviesLoader.load_movie_images(annotation_movie_path, self.framerate, self.selector)
                for annotation_id in annotation_frames['images']:
                    annotation_path = f"{annotation_movie_path}{annotation_id}"
                    annotation_id = f"{movie_id}{annotation_id}"
                    self.annotation_paths[annotation_id] = annotation_path

            # TODO assert that there is no inconsistency in images vs annotations movies: #frames, #framerate

        self.input_order = sorted(self.input_paths.keys())

    def list_movies(self):
        return self.files_loader.list_images()

    def list_movies_paths(self):
        return self.files_loader.list_images_paths()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.close()

    def close(self):
        if self.video_image_reader:
            self.video_image_reader.close()
            self.video_image_reader = None
        if self.video_annotation_reader:
            self.video_annotation_reader.close()
            self.video_annotation_reader = None

    @staticmethod
    def load_movie_images(movie_path, framerate: t.Optional[float], selector: FrameSelector) -> dict:
        images = []
        tags = []
        with StreamReader(movie_path) as video_reader:
            original_frame_rate = video_reader.frame_rate
            framerate = framerate or original_frame_rate
            samples = list(video_reader.pos_samples(framerate))

            for clip, clip_nr in selector.select(samples, original_frame_rate):
                images += [f"_{c:05}" for c in clip]
                for i_c, c in enumerate(clip):
                    tag = {"id": f"_{c:05}", "pos": c, "pos_clip": i_c, "clip_nr": clip_nr,
                           "timestamp": c * video_reader.frame_interval}
                    tags.append(tag)
        return {'images': images, 'tags': tags}

    def list_images_paths(self):
        return [self.input_paths[p] for p in self.input_order]

    def list_images(self):
        return list(self.input_order)

    def path_to_id(self, path):
        return path.stem  # TODO this may not be unique

    def split_frame_path(self, frame_path):
        """
        This extract movie and frame components from full frame_path.
        Args:
            frame_path: complete path to the frame: movies_path_frame_id

        Returns:
            path_to_movie, frame_id
        """
        path_to_movie, frame_id = frame_path.rsplit("_", maxsplit=1)
        return path_to_movie, frame_id

    def __get_frame_path(self, path_set, name_or_num):
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        if isinstance(name_or_num, str):
            return path_set.get(name_or_num, None)
        else:
            raise NotImplemented(type(name_or_num))

    @staticmethod
    def prepare_reader(video_reader, path_to_movie):
        if video_reader is not None and video_reader.input_string != path_to_movie:
            video_reader.close()
            video_reader = None
        if video_reader is None:
            video_reader = StreamReader(path_to_movie)
            video_reader.__enter__()
        return video_reader

    def load_image(self, name_or_num) -> np.ndarray:
        path_to_frame = self.__get_frame_path(self.input_paths, name_or_num)
        if path_to_frame is None:
            raise Exception(f"{name_or_num} does not exist in the loader.")
        path_to_movie, frame_nr = self.split_frame_path(path_to_frame)
        self.video_image_reader = MoviesLoader.prepare_reader(self.video_image_reader, path_to_movie)
        return self.video_image_reader[int(frame_nr)]

    def load_tag(self, name_or_num):
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        return self.json_tags.get(name_or_num, None)

    def load_annotation(self, name_or_num) -> t.Optional[pathlib.Path]:
        path_to_frame = self.__get_frame_path(self.annotation_paths, name_or_num)
        if path_to_frame is None:
            return None
        path_to_movie, frame_nr = self.split_frame_path(path_to_frame)
        self.video_annotation_reader = MoviesLoader.prepare_reader(self.video_annotation_reader, path_to_movie)
        # TODO this can be single or multichannel - perhaps someone should check that or convert to single channel?
        annotation_array = self.video_annotation_reader[int(frame_nr)]
        # The annotation video can be compressed so that it is not [0,1] or [0,255] but be a spread.
        if self.annotation_as_mask:
            return annotation_array >= 128
        return annotation_array

    def get_relative_path(self, name_or_num):
        """
        Determine the relative path to the given frame.
        Used by Savers.
        Args:
            name_or_num: frame_id or number in entire MoviesLoader

        Returns: movie_id/frame_id
        """
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        movie_id = self.load_tag(name_or_num)['movie_id']
        # TODO in perfect world it would do entire hierarchy up to the movie id
        return os.path.join(movie_id, name_or_num)

    def __str__(self):
        return f"MovieLoader for: {self.data_root}"



import unittest

import numpy as np
import numpy.testing as nptest

import sep.producers
import tests.testbase


class TestProducer(tests.testbase.TestBase):
    class Threshold(sep.producers.ImagesProducer):
        def segmentation(self, image, image_tag):
            return image > 128

    def test_producer_calculate(self):
        random_soup = np.random.random((10, 10)) / 2.0
        random_soup[3:5, 6:9] = 1

        random_soup_tags = {"source": "dummies", "id": "soup"}

        thresh_producer = TestProducer.Threshold("DummyThreshold", cache_root=self.create_temp_dir())
        segm, producer_tags = thresh_producer.calculate(random_soup, random_soup_tags)

        self.assertEqual(2 * 3, segm.sum())
        self.assertIn("run_time", producer_tags)
        self.assertIn("run_fps", producer_tags)
        self.assertIn("producer_details", producer_tags)
        self.assertIn("DummyThreshold", producer_tags["producer_details"])
        # self.assertIn('producer-hash', producer_tags)

    def test_producer_cache(self):
        random_soup = np.random.random((10, 10)) / 2.0
        random_soup_tags = {"source": "dummies", "id": "soup"}

        thresh_producer = TestProducer.Threshold("DummyThreshold", cache_root=self.create_temp_dir())
        segm, producer_tags = thresh_producer.calculate(random_soup, random_soup_tags)

        nptest.assert_equal(segm, thresh_producer.load_segment(random_soup_tags["id"]))
        self.assertEqual(producer_tags, thresh_producer.load_tag(random_soup_tags["id"]))


if __name__ == '__main__':
    unittest.main()



import fire
import os
import pathlib
import shutil

inspect_dir_path = pathlib.Path(__file__).parent
inspect_dir_name = os.path.basename(inspect_dir_path)


def _jupyter(inspect_dir):
    command = f'cd {inspect_dir}'
    print("Starting jupyter inside:", command)
    os.system(command)
    os.chdir(inspect_dir)
    command = f'jupyter notebook'
    os.system(command)


def here():
    _jupyter(inspect_dir_path)


def from_copy(dir_to_copy_inspect_to):
    os.makedirs(dir_to_copy_inspect_to, exist_ok=True)
    shutil.copytree(pathlib.Path(__file__).parent, os.path.join(dir_to_copy_inspect_to, inspect_dir_name))
    _jupyter(dir_to_copy_inspect_to)


if __name__ == '__main__':
    fire.Fire()



from sep._commons.imgutil import *
from tests.testbase import TestBase


class TestImgUtil(TestBase):
    def test_make_2d(self):
        with self.assertRaises(ValueError):
            image_4d = np.random.random((2, 2, 2, 2))
            make_2d(image_4d)

        image_2d = self.random_uint((20, 20))
        self.assertEqual(2, image_2d.ndim)
        res = make_2d(image_2d)
        self.np_assert_equal(res, image_2d)

        image_2d_rgb = make_rgb(image_2d)
        self.assertEqual(3, image_2d_rgb.ndim)
        res = make_2d(image_2d_rgb)
        self.np_assert_equal(res, image_2d)

        image_3d = self.random_rgb((20, 20))
        res = make_2d(image_3d, strict_duplication=False)  # this is possible, resulting is mean
        self.np_assert_equal(res.ndim, 2)

        with self.assertRaises(ValueError):
            make_2d(image_3d, strict_duplication=True)




from .files import FilesLoader
from .images import ImagesLoader
from .loader import Loader
from .movies import MoviesLoader, FrameByGroupSelector, FrameByIntervalSelector
from .youtube import YoutubeLoader



import typing

import numpy as np
import pandas as pd

from sep.assessors.metrics import Metric
from sep.assessors.regions import Region, EntireRegion


class Metricer:
    """
    This is class responsible for generating a set of metrics for various image regions.
    """
    metrics: typing.List[Metric]
    regions: typing.List[Region]

    def __init__(self):
        self.metrics = []
        self.regions = [EntireRegion()]
        self.reports = []

    def reset(self):
        self.reports = []

    def calculate_metrics(self, segmentation: np.ndarray, ground_truth: np.ndarray) -> pd.DataFrame:
        reports = []
        proper_report_columns_order = [metric.name for metric in self.metrics]
        proper_report_columns_order.insert(0, "region")

        for region in self.regions:
            seg_region = region.regionize(ground_truth=ground_truth, mask=segmentation)
            gt_region = region.regionize(ground_truth=ground_truth, mask=ground_truth)

            metrics_region = {metric.name: metric.calculate(seg_region, gt_region) for metric in self.metrics}
            region_report = pd.DataFrame.from_records([metrics_region])
            region_report["region"] = region.name
            reports.append(region_report)
        return pd.concat(reports).reindex(proper_report_columns_order, axis=1)

    def report_overall(self):
        """
        This should aggregate all the collected results.
        result_sampler per image region
            id, *img_tag, *seg_tag, iou, recall..., region_eval_path

        result_sample per image:
            id, name, *img_tag, *seg_tag, metrics (region_A_iou, region_A_path?, region_B_iou ..., iou_avg)

        result_sample per grouping:
            group_name, *seg_tag_avg, general_path_to_details??, region_A_iou_avg, region_B_iou_avg , iou_avg

        result_sample per entire run (single line):
            segmentator_name, *seg_tag_avg, region_A_iou_avg, region_B_iou_avg, iou_avg

        Returns:

        """
        # TODO multiple levels of aggregation
        return pd.concat(self.reports).fillna('').groupby(['region']).mean()

    def report_full(self):
        return pd.concat(self.reports).fillna('').reset_index(drop=True)

    def evaluate_image(self, image: np.ndarray, tag: dict, segment: np.ndarray, segment_tag: dict,
                       gt: np.ndarray) -> pd.DataFrame:

        """
        Evaluate the given image, ground truth and segmentation and store and aggregate the metrics report.
        """
        metric_report = self.calculate_metrics(segmentation=segment, ground_truth=gt)
        metric_report.insert(0, "id", tag["id"])
        # TODO multilayer tagging
        for (k, v) in tag.items():
            if k != "id":
                metric_report["img_" + k] = v

        for (k, v) in segment_tag.items():
            metric_report["seg_" + k] = v

        self.reports.append(metric_report)
        return metric_report



import numpy as np
import os
import sys
from IPython.display import Markdown, display, HTML
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap, NoNorm
from pathlib import Path

random_colormap = ListedColormap([(0, 0, 0)] + list(np.random.rand(20000, 3)))


def add_root_to_path(steps_up):
    root_dir = Path(__file__)
    for _ in range(steps_up):
        root_dir = root_dir.parent

    if root_dir not in sys.path:
        sys.path.append(str(root_dir))
    return root_dir


def pick_path(paths):
    correct_paths = [p for p in paths if os.path.exists(p)]
    if len(correct_paths) > 1:
        raise Exception("More than one existing path to choose.")
    elif len(correct_paths) == 0:
        raise Exception("No path exists.")
    return correct_paths[0]


def info(array, name=""):
    print(name, type(array))
    info_string = f"Shape: {array.shape}, Type: {array.dtype}, Min-max: {array.min(), array.max()}"
    info_string += f", Mean: {array.mean()}"
    print(info_string)


def show(*args, **kwargs):
    show_all(1, len(args), *args, **kwargs)


def show_all(rows, cols, *args, display_now=True, **kwargs):
    """
    Presents given arrays in form of grid as matplotlib figure.
    Args:
        rows: number of rows to show.
        cols: number of columns to show.
        *args: list of numpy arrays to show
        display_now: should the figure be displayed immediately
        **kwargs:
            - scale: define figsize
            - cmap: value mapping for each or every array
                if 'rand' then no normalization and random int to RGB mapping used
                else it is just passed to imshow
    Returns:
        matplotlib figure to show in notebook
    """
    scale = 30
    if 'scale' in kwargs:
        scale = kwargs['scale']

    subtitles = kwargs.get("titles", [""])

    gray_cmaps = ['gray']
    random_map = random_colormap
    normalization = [True]
    if 'cmap' in kwargs:
        gray_cmap_params = kwargs['cmap'].split(',')
        gray_cmaps = []
        normalization = []
        for m in gray_cmap_params:
            is_rand = m.strip() == 'rand'
            normalization.append(not is_rand)
            if is_rand:
                gray_cmaps.append(random_map)
            else:
                gray_cmaps.append(m)

    shape_ratio = args[0].shape[0] / args[0].shape[1]
    fig, axes = plt.subplots(rows, cols, figsize=(scale * cols, scale * rows * shape_ratio))
    fig.subplots_adjust(wspace=0.01, hspace=0.01)
    plt.subplots_adjust(wspace=0.01, hspace=0.01)
    for i in range(rows * cols):
        array = args[i] if len(args) > i else np.zeros(args[0].shape)
        if rows == 1 and cols == 1:
            ax = axes
        else:
            ax = axes[i] if rows == 1 or cols == 1 else axes[i // cols][i % cols]

        sub_ix = i
        subtitle = ""
        if len(subtitles) <= i:
            sub_ix = i % cols

        if len(subtitles) > sub_ix:
            subtitle = subtitles[sub_ix]

        ax.set_aspect("auto")
        ax.set_title(subtitle)
        if len(array.shape) == 3:
            if array.dtype == np.uint32 or array.dtype == np.uint16:
                array = (array / np.iinfo(array.dtype).max * 255).astype(np.uint8)
            ax.imshow(array)
        else:
            map_id = i % len(gray_cmaps)
            if normalization[map_id]:
                ax.imshow(array, cmap=gray_cmaps[map_id])
            else:
                ax.imshow(array, norm=NoNorm(), cmap=gray_cmaps[map_id])

    plt.close(fig)
    if display_now:
        display(fig)
    return fig


def printmd(*text):
    display(Markdown(" ".join([str(t) for t in text])))


def display_width(size_perc):
    display(HTML("<style>.container { width:{0}% !important; }</style>".format(size_perc)))



import traceback

import cv2
import numpy as np
import pathlib
import typing as t

from sep._commons import imgutil
from sep._commons.utils import *


class StreamReader:
    def __init__(self, input_string):
        self.input_string = str(input_string)
        self.frame_num = None
        self.current_frame = 0

    def __iter__(self):
        return self.read_samples(self.frame_rate)

    def read_samples(self, sampling_framerate: float) -> t.Iterable[np.ndarray]:
        current_time = 0
        next_sample_time = 0
        sampling_interval = 1.0 / sampling_framerate
        self.reader.set(cv2.CAP_PROP_POS_FRAMES, 0)
        for i in range(self.frame_num):
            res, image = self.reader.read()
            if not res:
                print(f"Failed to get frame nr {i}.")
            if next_sample_time <= current_time:
                yield image[..., ::-1]  # make it RGB
                next_sample_time += sampling_interval
            current_time += self.frame_interval

    def pos_samples(self, sampling_framerate: float) -> t.List[int]:
        res = []
        current_time = 0
        next_sample_time = 0
        sampling_interval = 1.0 / sampling_framerate
        for i in range(self.frame_num):
            if next_sample_time <= current_time:
                res.append(i)
                next_sample_time += sampling_interval
            current_time += self.frame_interval
        return res

    def move_to_item(self, item):
        if self.current_frame > item or item > self.current_frame + 500:
            self.reader.set(cv2.CAP_PROP_POS_FRAMES, item)
        else:
            while self.current_frame < item:
                self.reader.grab()
                self.current_frame += 1

    def __getitem__(self, item):
        self.move_to_item(item)
        res, image = self.reader.read()
        if not res:
            print(f"Failed to get frame nr {item}.")
        self.current_frame = item + 1
        return image[..., ::-1]

    @property
    def shape(self):
        return self.frame_h, self.frame_w

    def __len__(self):
        return self.frame_num

    def __enter__(self):
        assert self.input_string.startswith("http") or os.path.isfile(self.input_string), \
            f"File {self.input_string} does not exist."
        self.reader = cv2.VideoCapture(self.input_string)
        self.frame_rate = int(self.reader.get(cv2.CAP_PROP_FPS))
        self.frame_interval = 1.0 / self.frame_rate
        self.frame_h = int(self.reader.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.frame_w = int(self.reader.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.frame_num = int(self.reader.get(cv2.CAP_PROP_FRAME_COUNT))
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.close()

    def close(self):
        if self.reader is not None:
            self.reader.release()


class VideoWriter:
    def __init__(self, output_path: t.Union[str, pathlib.Path], output_shape=None, fps=None,
                 related_reader: StreamReader = None, encoding='mp4v'):
        """
        Args:
            output_path: path to the output file
            output_shape: (Y,X) shape of the imagery, ignored when related reader provided
            fps: frame rate of the output video, ignored when related reader provided
            related_reader: reader where the processed data will come from
            encoding: the desired encoding of the output file
        """
        assert related_reader is not None or \
               (len(output_shape) == 2 or len(output_shape) == 3 and output_shape[2] == 3,
                "Incorrect shape " + str(output_shape))

        self.encoding = encoding
        self.frames_num = 0
        if related_reader is None:
            assert fps is not None and output_shape is not None
            self.frame_rate = fps
            self.frame_h = output_shape[0]
            self.frame_w = output_shape[1]
        else:
            self.frame_rate = int(related_reader.frame_rate)
            self.frame_h, self.frame_w = related_reader.shape
        self.output_path = str(output_path)

    def add(self, image: np.ndarray, accept_bool=False, accept_float=False):
        """
        Args:
            image: RGB or single channel image
            accept_float: convert float 0-1 to uint8
            accept_bool: convert bool to uint8
        """
        assert accept_bool or not image.dtype == bool
        assert accept_float or not image.dtype == float
        assert_arg(image.shape[:2] == self.shape, f"image.shape not compatible with {self.shape}")

        bgr_array = imgutil.make_rgb(image)[..., ::-1]
        self.video_writer.write(bgr_array)

    @property
    def shape(self):
        return self.frame_h, self.frame_w

    def __enter__(self):
        self.video_writer = cv2.VideoWriter(self.output_path,
                                            cv2.VideoWriter_fourcc(*self.encoding),
                                            self.frame_rate,
                                            (self.frame_w, self.frame_h))
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.close()

    def close(self):
        self.video_writer.release()



import os
import unittest

import numpy as np
import numpy.testing as nptest

from sep._commons.utils import *
from sep.loaders import ImagesLoader
from sep.savers.images import ImagesSaver
from tests.testbase import TestBase


class TestImagesSaver(TestBase):
    def test_saving_in_hierarchy(self):
        test_images_loader = ImagesLoader.from_tree(self.root_test_dir("input/basics"))
        temp_dir = self.create_temp_dir()

        saver = ImagesSaver(temp_dir, test_images_loader)
        names = test_images_loader.list_images()
        self.assertEqual(7, len(names))
        self.assertEqual("human_1", names[0])
        known_name = "lights01"

        sample_res_1 = self.random_rgb((20, 20))
        sample_tag_1 = {"id": "human_1", "detail": 20}
        saver.save_result(0, sample_res_1)
        saver.save_tag(0, sample_tag_1, result_tag={"seg_confident": "no"})
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, 'humans', 'human_1.png')))
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, 'humans', 'human_1.json')))

        sample_res_2 = self.random_rgb((20, 20))
        saver.save_result(known_name, sample_res_2)
        self.assertTrue(os.path.isfile(os.path.join(temp_dir, 'lights', 'lights01.png')))

        saved_images_loader = ImagesLoader.from_tree(temp_dir)
        names = saved_images_loader.list_images()
        self.assertEqual(2, len(names))

        humans_saved = saved_images_loader['human_1']
        nptest.assert_almost_equal(humans_saved['image'], sample_res_1)
        self.assertEqual(humans_saved['tag'], union(sample_tag_1, {"seg_confident": "no"}))

        lights_saved = saved_images_loader['lights01']
        nptest.assert_almost_equal(lights_saved['image'], sample_res_2)
        self.assertEqual(lights_saved['tag'], {'id': 'lights01'})  # defaults


if __name__ == '__main__':
    unittest.main()



import numpy as np

from sep._commons.utils import *


def get_2d_size(image: np.ndarray):
    assert 2 <= image.ndim <= 3
    if image.ndim == 3:  # assume it is multichannel with last dim as channels
        return image.shape[:-1]
    return image.shape


def mix_rgb(red: np.array, green: np.array, blue: np.array):
    shape = None
    if red is not None:
        assert_arg(red.ndim == 2, "red", "{} has to be 2d.")
        assert_value(shape is None or red.shape == shape, "Arrays have to have the same shape.")
        shape = red.shape
    if green is not None:
        assert_arg(green.ndim == 2, "green", "{} has to be 2d.")
        assert_value(shape is None or green.shape == shape, "Arrays have to have the same shape.")
        shape = green.shape
    if blue is not None:
        assert_arg(blue.ndim == 2, "blue", "{} has to be 2d.")
        assert_value(shape is None or blue.shape == shape, "Arrays have to have the same shape.")
        shape = blue.shape

    rgb = np.zeros(list(shape) + [3], dtype=np.uint8)
    for channel, array in enumerate([red, green, blue]):
        if array is not None:
            rgb[..., channel] = array
    return rgb


def make_rgb(image: np.ndarray):
    assert image.ndim >= 2
    if image.dtype == bool or np.issubdtype(image.dtype, np.floating) and image.max() <= 1:
        image = (image * 255)
    if image.dtype != np.uint8:
        image = image.astype(np.uint8)

    if image.ndim == 2:
        return np.stack([image, image, image], axis=-1)

    return image


def make_2d(image: np.ndarray, strict_duplication=False):
    if image.ndim == 2:
        return image
    elif image.ndim == 3:
        if strict_duplication:
            assert_value(np.all(image.mean(axis=2) == image[...,0]), "image has channels with non-equal values")
        return image.mean(axis=2)
    else:
        raise ValueError(f"No way to handle more that 3d data ({image.ndim}).")



import numpy as np
import numpy.testing as nptest
import unittest

from sep.assessors.metricer import Metricer
from sep.assessors.metrics import IouMetric
from sep.assessors.regions import Region, EdgesRegion, RegionExpr
from tests.testbase import TestBase


class TestMetrics(TestBase):
    def test_iou(self):
        iou = IouMetric()
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        metric = iou.calculate(blob_2, blob_1)
        self.assertAlmostEqual(5.0 / (50 + 5), metric, places=5)


class TestMetricer(TestBase):
    class DummyRegion(Region):
        def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
            return ground_truth.astype(bool)

    def test_basic_metricer(self):
        metricer = Metricer()
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        empty_report = metricer.calculate_metrics(blob_2, blob_1)
        self.assertEqual(0, len(empty_report))

        iou_metric = IouMetric()
        metricer.metrics.append(iou_metric)
        report = metricer.calculate_metrics(blob_2, blob_1)
        self.assertEqual(1, len(report))
        nptest.assert_almost_equal([5.0 / (50 + 5)], report[iou_metric.name].values)
        nptest.assert_equal(["Entire image"], report["region"].values)

    def test_two_regions_metricer(self):
        metricer = Metricer()
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        iou_metric = IouMetric()
        new_region = self.DummyRegion(name="Dummy")
        metricer.metrics.append(iou_metric)
        metricer.regions.append(new_region)

        report = metricer.calculate_metrics(blob_2, blob_1)
        self.assertEqual(2, len(report))
        nptest.assert_almost_equal([5.0 / (50 + 5), 5.0 / 50], report[iou_metric.name].values)
        nptest.assert_equal(["Entire image", "Dummy"], report["region"].values)

    def test_downscale_region_metricer(self):
        blob_rand_bin_1 = np.random.random((1080, 1920)) > 0.5
        blob_rand_bin_2 = np.random.random((1080, 1920)) > 0.5

        metricer = Metricer()
        metricer.metrics.append(IouMetric())
        metricer.regions.append(RegionExpr('~', EdgesRegion(0.02, downsample_x=640), name="Mask robust"))
        metricer.calculate_metrics(blob_rand_bin_1, blob_rand_bin_2)

    def test_evaluate_image(self):
        metricer = Metricer()
        metricer.metrics.append(IouMetric())

        image = np.random.random((10, 10, 3))
        gt = np.zeros((10, 10), dtype=bool)
        gt[0:5, 0:10] = True
        seg = np.zeros((10, 10))
        seg[4:6, 0:5] = 1

        report = metricer.evaluate_image(image, tag={"id": "blobix"}, gt=gt,
                                         segment=seg, segment_tag={"name": "Resnet", "fps": 20})
        self.assertEqual(1, len(report))
        self.assertEqual(5, len(report.columns))
        nptest.assert_almost_equal(report["iou"].values, [5.0 / (50 + 5)])
        self.assertEqual(report["id"].values, ["blobix"])
        self.assertEqual(report["region"].values, ["Entire image"])
        self.assertEqual(report["seg_name"].values, ["Resnet"])
        self.assertEqual(report["seg_fps"].values, [20])


if __name__ == '__main__':
    unittest.main()



import fire
import shutil
from tqdm import tqdm

import sep.loaders
from sep._commons.utils import *
from sep.loaders.loader import Loader
from sep.savers.saver import Saver


def extract_to_images(data_loader: Loader, data_saver: Saver,
                      output_root,
                      remove_existing=False,
                      verbose=1):
    if verbose:
        print(f"Extracting images from {data_loader}.")
        print(f"Results will be saved at {output_root} using {data_saver}.")
        print(f"There are {len(data_loader)} images to process.")

    data_saver.set_output(output_root, data_loader)
    if remove_existing and os.path.exists(output_root):
        shutil.rmtree(output_root)
    os.makedirs(output_root)

    for i in tqdm(range(len(data_loader)), "Extracting images"):
        image = data_loader.load_image(i)
        tag = data_loader.load_tag(i)

        data_saver.save_result(i, image)
        data_saver.save_tag(i, tag, result_tag={})
    data_saver.close()

    extracted_loader = sep.loaders.ImagesLoader.from_tree(output_root)
    if verbose:
        print(f"Extracted {len(data_loader)} images and loader found {len(extracted_loader)} of them.")
    if len(data_loader) != len(extracted_loader):
        print("Different count of the extracted and reloaded images!")
    return extracted_loader


def extract_from_movies_to_images(movies_root, output_root):
    movies_loader = sep.loaders.MoviesLoader(movies_root, framerate=None, clips_skip=0, clips_len=10000000, verbose=1)
    saver = sep.savers.ImagesSaver(output_root, movies_loader)
    extract_to_images(movies_loader, saver, output_root)


def extract_from_youtube_to_images(url_list_path, output_root, quality="720p"):
    """
    Extract image frames from youtube videos defined in the url_list_path file.
    File format is:
    <url>;<start_time>;<end_time>;<frames_numbers>;<optional_tags>
    <start_time> / <end_time> = <min>:<sec>
    <optional_tags> = <tag_name>:<tag_value>;<optional_tags>
    Args:
        url_list_path: path to the file with youtube videos to extract
        output_root: directory where the image frames will be saved
        quality: string description
    """
    url_samples = [l.strip() for l in open(url_list_path, "r").readlines()]

    # Parse url sample.
    extract_samples = []
    for url_sample in url_samples:
        tokens = [s.strip() for s in url_sample.split(";")]
        sample = {'url': tokens[0], 'start_time': tokens[1], 'end_time': tokens[2], 'frames_numbers': int(tokens[3])}

        # parse tags
        tags = {}
        for t in tokens[4:]:
            key, value = map(str.strip, t.split(":"))
            tags[key] = value

        sample['start_time'] = time_to_sec(sample['start_time'])
        sample['end_time'] = time_to_sec(sample['end_time'])
        sample['tags'] = tags
        extract_samples.append(sample)

    # Saving from youtube.
    saver = sep.savers.ImagesSaver()
    with sep.loaders.YoutubeLoader(quality) as youtube_loader:
        for sample in tqdm(extract_samples, "Preparing extraction plan"):
            url = sample['url']
            selector = sep.loaders.FrameByIntervalSelector(frames_to_extract=sample['frames_numbers'],
                                                           interval_start_time=sample['start_time'],
                                                           interval_end_time=sample['end_time'])
            youtube_loader.add_url(url, selector, user_info=sample['tags'])
        extract_to_images(youtube_loader, saver, output_root, remove_existing=True)


if __name__ == '__main__':
    fire.Fire()



import json
import os
import pathlib


def update_with_suffix(to_update: dict, new_items: dict, prefix: str = '', suffix: str = '') -> dict:
    new_items = {prefix + k + suffix: v for k, v in new_items.items()}
    to_update.update(new_items)
    return to_update


def add_defaults(to_update: dict, defaults: dict):
    for k, v in defaults.items():
        if k not in to_update:
            to_update[k] = v
    return to_update


def get_with_prefix(dict_with_prefix: dict, prefix: str = '', strip_prefix=True) -> dict:
    return {k[len(prefix):] if strip_prefix else k: v for k, v in dict_with_prefix.items() if k.startswith(prefix)}


def subset(base_dictionary: dict, keys_subset):
    keys_subset = set(keys_subset)
    return {k: v for k, v in base_dictionary.items() if k in keys_subset}


def union(*data: dict):
    return {k: v for d in data for k, v in d.items()}


def load_json(path_to_file):
    with open(str(path_to_file), 'r') as f:
        return json.load(f)


def save_json(path, *data: dict):
    os.makedirs(os.path.dirname(str(path)), exist_ok=True)
    with open(str(path), 'w') as f:
        json.dump(union(*data), f, indent=4, sort_keys=True)


def time_to_sec(time: str) -> int:
    time_tokens = [int(t) for t in time.split(":")]
    res = 0
    for v in time_tokens:
        res = res * 60 + v
    return res


def assert_arg(arg_assert, arg_name, format_message="Invalid parameter: {}"):
    if not arg_assert:
        raise ValueError(format_message.format(arg_name))


def assert_value(value, description):
    if not value:
        raise ValueError(description)


def ensure_posix_path_str(path: str) -> str:
    if path is None:
        return None
    return str(pathlib.Path(pathlib.PureWindowsPath(path).as_posix()))


class SepException(Exception):
    pass



import os
import typing
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from sep.assessors.detailer import Detailer
from sep.assessors.metricer import Metricer
from sep.loaders import Loader
from sep.producers import Producer


def evaluate(data_loader: Loader, producer: Producer, metricer: Metricer,
             detailer: typing.Optional[Detailer], output_evalpath=None,
             verbose=1):
    if verbose:
        print(f"Evaluation of {producer} on data from {data_loader}.")
        print(f"There are {len(data_loader)} images to evaluate on.", flush=True)

    if output_evalpath is not None:
        os.makedirs(output_evalpath, exist_ok=True)

    for i in tqdm(range(len(data_loader)), producer.name):
        image = data_loader.load_image(i)
        gt = data_loader.load_annotation(i)
        tag = data_loader.load_tag(i)

        if gt is None:
            print(data_loader.list_images()[i], "does not have annotation data!")

        segment, segment_tag = producer.calculate(image, tag)
        data_point_eval = metricer.evaluate_image(image, tag, segment, segment_tag, gt)
        if detailer is not None:  # TODO detailer vs metricer relation has to be rethought
            detailer.save_image_evaluation(data_point_eval)

    return metricer.report_overall()


def compare(data_loader: Loader, producers: typing.List[Producer], metricer, detailer, output_evalpath):
    print(f"Comparison of {len(producers)} producers on data from {data_loader}.", flush=True)

    reports = []
    for producer in producers:
        if output_evalpath is not None:
            producer_eval_path = Path(output_evalpath) / producer.name
        else:
            producer_eval_path = None
        producer_report = evaluate(data_loader, producer, metricer, detailer, producer_eval_path,
                                   verbose=0)
        producer_report.insert(loc=0, column='producer', value=producer.name)

        reports.append(producer_report)
    return pd.concat(reports).groupby(['producer', 'region']).mean()



import typing as t

import numpy as np
import skimage
from skimage.color.colorlabel import DEFAULT_COLORS

import sep.inspect.utils


def overlay_region(image: np.ndarray, region: np.ndarray):
    # TODO work not only on RGB image? # Check if it is a mask / label?
    overlay = skimage.color.label2rgb(region, image, colors=DEFAULT_COLORS, bg_label=0)
    return overlay


def overlay_regions(image: np.ndarray, regions: t.List[np.ndarray], regions_names: t.Optional[t.List[str]]):
    if regions_names is None:
        # use some dummy naming
        regions_names = [str(i) for i in range(1, len(regions) + 1)]
    assert len(regions) == len(regions_names)

    flatten_label = np.zeros(regions[0].shape, dtype=np.uint8)
    legend = []
    for label, (name, region) in enumerate(zip(regions_names, regions), start=1):
        flatten_label[region > 0] = label
        legend.append((name, DEFAULT_COLORS[label - 1]))

    overlay = skimage.color.label2rgb(flatten_label, image, bg_label=0)

    return overlay, legend


def show_regions_and_fractions(regions, annotation):
    res = []
    for region in regions:
        region_frac, anno_frac = region.area_fractions(annotation)
        desc = {'region frac:': region_frac, 'annotation_frac': anno_frac, 'viable': region.is_viable_for(annotation)}
        print(f"Result for {region} is: {desc}.")
        res.append(region.extract_region(annotation))
    sep.inspect.utils.show(*res)



import json
import pathlib
from abc import ABC, abstractmethod
from timeit import default_timer as timer

import imageio
import numpy as np

from sep._commons.utils import *
from sep.producers.producer import Producer


class ImagesProducer(Producer, ABC):
    """
    This is responsible for creating segmentation that will be later evaluated.
    It should be able to cache the results and add processing related tags.
    """
    def __init__(self, name, cache_root):
        # TODO add memory cache (lru based) and make cache root optional or replace it with some storage object
        super().__init__(name)
        self.cache_root = pathlib.Path(cache_root)

    def load_segment(self, id):
        cache_path = (self.cache_root / str(id)).with_suffix(".tif")
        return imageio.imread(str(cache_path))

    def load_tag(self, id):
        cache_path = (self.cache_root / str(id)).with_suffix(".json")
        return load_json(cache_path)

    def __save_segment(self, id, segm):
        cache_path = (self.cache_root / str(id)).with_suffix(".tif")
        imageio.imsave(str(cache_path), segm)

    def __save_tag(self, id, tag):
        cache_path = (self.cache_root / str(id)).with_suffix(".json")
        save_json(cache_path, tag)

    def calculate(self, input_image: np.ndarray, input_tag: dict) -> (np.ndarray, dict):
        # TODO is possible and requested load results and tags from cache
        assert input_tag is not None
        if input_image.dtype != np.uint8: # TODO we need to be sure
            input_image = (input_image * 255).astype(np.uint8)
        assert input_image.dtype == np.uint8

        seg, seg_tag = super().calculate(input_image, input_tag)
        # TODO what if the seg is not uint8? What if it is not mask? What if it is predictions as floats?
        if isinstance(seg, np.ndarray) and seg.dtype == bool:
            seg = seg.astype(np.uint8)

        if self.cache_root and "id" in input_tag:
            self.__save_tag(input_tag["id"], seg_tag)
            self.__save_segment(input_tag["id"], seg)

        return seg, seg_tag

    def __repr__(self):
        return f"{self.__class__} ({self.__dict__})"





import os
import unittest

from sep.loaders import MoviesLoader
from sep.savers import MoviesSaver
from tests.testbase import TestBase


class TestMoviesSaver(TestBase):
    def test_saving_in_hierarchy(self):
        with MoviesLoader.from_tree(self.root_test_dir("input"), framerate=5, clips_len=1, clips_skip=10) as movies_loader:
            temp_dir = self.create_temp_dir()
            movies = movies_loader.list_movies()
            self.assertEqual(3, len(movies))
            self.assertEqual(10, len(movies_loader.list_images()))
            self.assertEqual("Dragon - 32109", movies[1])
            movies_loader.load_image(0)

            with MoviesSaver(temp_dir, loader=movies_loader) as saver:
                sample_res_1 = self.random_rgb((360, 640))
                sample_tag_1 = {"movie_id": "dragon_1", "movie_source": "pixabay", "movie_type": "flying"}
                saver.save_result(0, sample_res_1)
                saver.save_tag(0, sample_tag_1, result_tag={'seg_name': 'fast'})
                saver.save_result(8, sample_res_1)
                saver.save_tag(8, sample_tag_1, result_tag={'seg_name': 'slow'})

            self.assertTrue(os.path.isfile(os.path.join(temp_dir, "Dinosaur - 1438.mp4")))
            self.assertTrue(os.path.isfile(os.path.join(temp_dir, "Dragon - 32109.mp4")))

        with MoviesLoader.from_tree(temp_dir, framerate=5, clips_len=1, clips_skip=0) as reload_movies:
            self.assertEqual(2, len(reload_movies.list_movies()))
            self.assertEqual(2, len(reload_movies.list_images()))
            # TODO check tag if it is stripped of the movie per movie file
            # TODO check tag read - it should be similar to the saved one


if __name__ == '__main__':
    unittest.main()



import skimage
from matplotlib import pyplot as plt, patches as mpatches
from skimage.color.colorlabel import DEFAULT_COLORS

import sep._commons.imgutil as imgutil


def overlay(image, labels):
    return skimage.color.label2rgb(labels, image, colors=DEFAULT_COLORS, bg_label=0)


def overlay_as_rgb(image, red, green, blue, pred_alpha=0.75):
    """
    Args:
        image: grayscale or rgb image
        red: red channel
        green: green channel
        blue: blue channel
        pred_alpha: image colour mapped image blending (1 is show only prediction)

    Returns:
        RGB np array
    """
    # TODO check 0-1 vs 0-255, for now assume masks
    # TODO this should be rewritter using basic imgutil and visual functions - it is just an overlay
    inspect = imgutil.mix_rgb(red, green, blue)

    inspect *= 255
    if image.ndim == 2:
        image = skimage.color.gray2rgb(image)

    return imgutil.make_rgb(pred_alpha * inspect + (1 - pred_alpha) * image)


def overlay_prediction(image, prediction, cmap='gnuplot2', pred_alpha=0.5):
    """
    Args:
        image: grayscale or rgb image
        prediction: float np array
        cmap: name of the cmap to use
        pred_alpha: image colour mapped image blending (1 is show only prediction)

    Returns:
        RGB np array
    """
    if image.ndim == 2:
        image = skimage.color.gray2rgb(image)
    cm = plt.get_cmap(cmap)
    prediction_rgb = cm(prediction)[..., 0:3] * 255
    return imgutil.make_rgb(pred_alpha * prediction_rgb + (1 - pred_alpha) * image)


def show_with_legend(images, legends, titles="", scale=None, legend_size=16):
    if not isinstance(images, list):
        images = [images]
    if not isinstance(legends, list):
        legends = [legends]
    elif not isinstance(legends[0], list):  # TODO it looks silly
        legends = [legends]
    if not isinstance(titles, list):
        titles = [titles]

    scale = scale or 30
    shape_ratio = images[0].shape[0] / images[0].shape[1]
    fig, axes = plt.subplots(1, len(images), figsize=(scale, scale * shape_ratio))
    if len(images) == 1:
        axes = [axes]

    for ax, image, legend, title in zip(axes, images, legends, titles):
        ax.set_aspect("auto")
        ax.set_title(title)
        legend_patches = [mpatches.Patch(color=colour, label=name) for name, colour in legend]
        ax.legend(handles=legend_patches, prop={'size': legend_size})
        ax.imshow(image)

    plt.close(fig)
    return fig



from sep._commons.utils import *
from tests.testbase import TestBase


class TestUtils(TestBase):
    def test_time_to_sec(self):
        self.assertEqual(15, time_to_sec("15"))
        self.assertEqual(75, time_to_sec("1:15"))
        self.assertEqual(75, time_to_sec("01:15"))
        self.assertEqual(7275, time_to_sec("02:01:15"))

    def test_pick_subset(self):
        dictionary = {'a': 1, 'b': 3, 'c': 4}
        res = subset(dictionary, ['a'])
        self.assertDictEqual(res, {'a': 1})
        res = subset(dictionary, ['a', 'c'])
        self.assertDictEqual(res, {'a': 1, 'c': 4})
        res = subset(dictionary, ['b', 'c'])
        self.assertDictEqual(res, {'b': 3, 'c': 4})



import pathlib

from abc import ABC, abstractmethod
import typing as t

import numpy as np


class Loader(ABC):
    def __init__(self):
        self.annotator = None
        pass

    @abstractmethod
    def list_images(self):
        pass

    @abstractmethod
    def load_image(self, name_or_num) -> t.Union[np.ndarray, pathlib.Path]:
        pass

    @abstractmethod
    def load_tag(self, name_or_num):
        pass

    @abstractmethod
    def load_annotation(self, name_or_num) -> t.Union[np.ndarray, pathlib.Path]:
        pass

    @abstractmethod
    def get_relative_path(self, name_or_num):
        pass

    def save_annotation(self, name_or_num, new_annotation, keep_history=False):
        pass

    def save_tag(self, name_or_num, new_tag):
        pass

    def extend_tag(self, name_or_num, tag_extensions: dict):
        current_tags = self.load_tag(name_or_num)
        current_tags.update(tag_extensions)
        self.save_tag(name_or_num, current_tags)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, item):
        return {"image": self.load_image(item),
                "tag": self.load_tag(item),
                "annotation": self.load_annotation(item)}

    def __len__(self):
        return len(self.list_images())

    def validate_annotation(self, annotation: np.ndarray):
        # TMP require it to be a mask
        assert annotation.ndim == 2



class Detailer:
    """
    This class is responsible for producing visual representation of the segmentation quality when confronted
    with ground truth.
    """
    def __init__(self, path, regions):
        pass

    def save_image_evaluation(self, data_point_eval):
        pass


import itertools
import numpy as np
import os
import unittest
from pathlib import Path

import sep.loaders.movies
from sep.loaders import MoviesLoader, FrameByGroupSelector, FrameByIntervalSelector
from tests.testbase import TestBase


class TestMoviesLoader(TestBase):
    def test_loading(self):
        with MoviesLoader.from_tree(self.root_test_dir("input/reptiles"),
                                    framerate=10, clips_len=5, clips_skip=10) as test_movies_loader:
            self.assertEqual(2, len(test_movies_loader.list_movies()))
            self.assertEqual(2, len(test_movies_loader.list_movies_paths()))

            self.assertEqual('Dinosaur - 1438_00000', test_movies_loader.list_images()[0])
            self.assertIn('Dragon - 32109_00000', test_movies_loader.list_images())
            self.assertEqual(63, len(test_movies_loader.list_images_paths()))

            # TODO check that first data
            tag_1 = test_movies_loader.load_tag('Dinosaur - 1438_00000')
            annotation_1 = test_movies_loader.load_annotation('Dinosaur - 1438_00000')
            image_1 = test_movies_loader.load_image('Dinosaur - 1438_00000')
            self.assertArray(image_1, 3, np.uint8)
            self.assertIsNone(annotation_1)
            self.assertSubset(tag_1, {'id': 'Dinosaur - 1438_00000', 'pos': 0, 'pos_clip': 0, 'clip_nr': 0, 'timestamp': 0.0})
            self.assertSubset(tag_1, {'movie_id': 'dinosaur_1', 'movie_source': 'pixabay', 'movie_type': 'running'})

            tag_2 = test_movies_loader.load_tag('Dinosaur - 1438_00037')
            annotation_2 = test_movies_loader.load_annotation('Dinosaur - 1438_00037')
            image_2 = test_movies_loader.load_image('Dinosaur - 1438_00037')
            self.assertArray(image_2, 3, np.uint8)
            self.assertIsNone(annotation_2)
            self.assertSubset(tag_2, {'id': 'Dinosaur - 1438_00037', 'pos': 37, 'pos_clip': 1, 'clip_nr': 1})
            self.assertAlmostEqual(1.608, tag_2['timestamp'], places=2)
            self.assertSubset(tag_2, {'movie_id': 'dinosaur_1', 'movie_source': 'pixabay', 'movie_type': 'running'})

            next_image_id = test_movies_loader.list_images().index('Dinosaur - 1438_00037') + 1

            tag_3 = test_movies_loader.load_tag(next_image_id)
            self.assertSubset(tag_3, {'id': 'Dinosaur - 1438_00040', 'pos': 40, 'pos_clip': 2, 'clip_nr': 1})
            self.assertAlmostEqual(1.739, tag_3['timestamp'], places=2)

            # images different
            self.np_assert_not_equal(image_1, image_2)

            # load last image
            tag_last = test_movies_loader.load_tag(62)
            annotation_last = test_movies_loader.load_annotation(62)
            image_last = test_movies_loader.load_image(62)
            self.assertArray(image_last, 3, np.uint8)
            self.assertIsNone(annotation_last)
            self.assertSubset(tag_last, {'id': 'Dragon - 32109_00311', 'pos': 311, 'pos_clip': 2, 'clip_nr': 7})
            self.assertAlmostEqual(10.724, tag_last['timestamp'], places=2)
            self.assertSubset(tag_last, {'movie_id': 'dragon_1', 'movie_source': 'pixabay', 'movie_type': 'flying'})

    def test_loading_iterations(self):
        with MoviesLoader.from_tree(self.root_test_dir("input/reptiles"),
                                    framerate=5, clips_len=1, clips_skip=10) as test_movies_loader:
            self.assertEqual(9, len(test_movies_loader.list_images_paths()))
            for sample in itertools.islice(test_movies_loader, 0, 5):
                tag = sample['tag']
                annotation = sample['annotation']
                image = sample['image']
                self.assertArray(image, 3, np.uint8)
                self.assertIsNone(annotation)
                self.assertSubset(tag, {'id': 'Dinosaur - 1438_00000', 'pos': 0, 'pos_clip': 0, 'clip_nr': 0})
                self.assertAlmostEqual(0, tag['timestamp'], places=2)
                self.assertSubset(tag, {'movie_id': 'dinosaur_1', 'movie_source': 'pixabay', 'movie_type': 'running'})
                return

    def test_load_movie_images(self):
        frame_selector = sep.loaders.movies.FrameByGroupSelector(clips_len=100000, clips_skip=30)
        loaded_frames = MoviesLoader.load_movie_images(self.root_test_dir("input/reptiles/Dinosaur - 1438.mp4"),
                                                       framerate=30, selector=frame_selector)
        loaded_images = loaded_frames['images']
        loaded_tags = loaded_frames['tags']
        self.assertEqual(157, len(loaded_images))
        self.assertEqual(157, len(loaded_tags))

        loaded_frames = MoviesLoader.load_movie_images(self.root_test_dir("input/reptiles/Dinosaur - 1438.mp4"),
                                                       framerate=None, selector=frame_selector)
        self.assertEqual(157, len(loaded_frames['images']))

        frame_selector = sep.loaders.movies.FrameByGroupSelector(clips_len=2, clips_skip=10)
        loaded_frames = MoviesLoader.load_movie_images(self.root_test_dir("input/reptiles/Dinosaur - 1438.mp4"),
                                                       framerate=None, selector=frame_selector)
        loaded_tags = loaded_frames['tags']
        self.assertEqual(27, len(loaded_tags))
        self.assertEqual('_00001', loaded_tags[1]['id'])
        self.assertEqual(1, loaded_tags[1]['pos'])
        self.assertEqual(0, loaded_tags[1]['clip_nr'])
        self.assertEqual('_00012', loaded_tags[2]['id'])
        self.assertEqual(12, loaded_tags[2]['pos'])
        self.assertEqual(1, loaded_tags[2]['clip_nr'])

    def test_relative(self):
        with MoviesLoader.from_tree(self.root_test_dir("input/reptiles"),
                                    framerate=5, clips_len=1, clips_skip=10) as movies_loader:
            data_names = movies_loader.list_images()
            self.assertEqual(9, len(data_names))
            self.assertEqual("Dinosaur - 1438_00000", data_names[0])
            self.assertEqual(os.path.join("dinosaur_1", "Dinosaur - 1438_00000"), movies_loader.get_relative_path(0))
            self.assertEqual(os.path.join("dinosaur_1", "Dinosaur - 1438_00000"),
                             movies_loader.get_relative_path("Dinosaur - 1438_00000"))

    def test_movie_annotations(self):
        with MoviesLoader.from_tree(self.root_test_dir("input/fafa"), framerate=5) as movies_loader:
            self.assertEqual(11, len(movies_loader))
            sample_0 = movies_loader[0]
            self.assertEqual((720, 1080, 3), sample_0['image'].shape)
            self.assertEqual((720, 1080, 3), sample_0['annotation'].shape)
            # it is compressed so that it is not exactly [0,1] or [0,255]
            self.assertTrue(len(np.unique(sample_0['annotation'])) > 60)
            sample_8 = movies_loader[8]
            self.assertEqual((720, 1080, 3), sample_8['image'].shape)
            self.assertEqual((720, 1080, 3), sample_8['annotation'].shape)
            self.assertTrue(len(np.unique(sample_0['annotation'])) > 60)
            self.np_assert_not_equal(sample_0['annotation'], sample_8['annotation'])

        # now with 'make-it-a-mask' option on
        with MoviesLoader.from_tree(self.root_test_dir("input/fafa"), framerate=5, annotation_as_mask=True) as movies_loader:
            sample_0 = movies_loader[0]
            self.assertEqual((720, 1080, 3), sample_0['annotation'].shape)
            # now it is turned into mask
            self.np_assert_equal([False, True], np.unique(sample_0['annotation']))

    def test_loading_with_listing(self):
        with self.create_temp("loader_listing.txt") as listing_file:
            listing_file.writelines(f"{Path('reptiles/Dragon - 32109.mp4')}, {Path('reptiles/Dragon - 32109_gt.mp4')}\n")

        with MoviesLoader.from_listing(self.root_test_dir("input"), filepath="loader_listing.txt",
                                       framerate=10, clips_len=5, clips_skip=10) as test_movies_loader:
            self.assertEqual(1, len(test_movies_loader.list_movies()))
            self.assertEqual(1, len(test_movies_loader.list_movies_paths()))

            tag = test_movies_loader.load_tag(0)
            self.assertEqual('Dragon - 32109', tag['movie_id'])
            self.assertIsNone(test_movies_loader.load_annotation(0))

        # now with tag
        with self.create_temp("loader_listing.txt") as listing_file:
            listing_file.writelines(
                f"{Path('reptiles/Dragon - 32109.mp4')}, {Path('reptiles/Dragon - 32109_gt.mp4')}, {Path('reptiles/Dragon - 32109.json')}\n")

        with MoviesLoader.from_listing(self.root_test_dir("input"), filepath="loader_listing.txt",
                                       framerate=10, clips_len=5, clips_skip=10) as test_movies_loader:
            self.assertEqual(1, len(test_movies_loader.list_movies()))

            tag = test_movies_loader.load_tag(0)
            self.assertEqual('dragon_1', tag['movie_id'])
            self.assertEqual('pixabay', tag['movie_source'])
            self.assertIsNone(test_movies_loader.load_annotation(0))


class TestFrameSelector(TestBase):
    def test_frame_group_selector(self):
        data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        selector = FrameByGroupSelector(clips_len=2, clips_skip=1, skip_first=0, skip_last=0)
        self.assertEqual([([1, 2], 0), ([4, 5], 1), ([7, 8], 2), ([10], 3)], list(selector.select(data)))
        selector = FrameByGroupSelector(clips_len=2, clips_skip=1, skip_first=1, skip_last=3)
        self.assertEqual([([2, 3], 0), ([5, 6], 1)], list(selector.select(data)))

    def test_frame_interval_selector(self):
        data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        selector = FrameByIntervalSelector(frames_to_extract=1, interval_start_time=1, interval_end_time=10)
        self.assertEqual([([6], 0)], list(selector.select(data)))

        selector = FrameByIntervalSelector(frames_to_extract=20, interval_start_time=1, interval_end_time=10)
        self.assertEqual(20, len(list(selector.select(data))))  # let it repeat itself

        selector = FrameByIntervalSelector(frames_to_extract=2, interval_start_time=1, interval_end_time=10)
        self.assertEqual([([4], 0), ([7], 1)], list(selector.select(data)))

        selector = FrameByIntervalSelector(frames_to_extract=3, interval_start_time=1, interval_end_time=10)
        self.assertEqual([([3], 0), ([6], 1), ([8], 2)], list(selector.select(data)))

        selector = FrameByIntervalSelector(frames_to_extract=4, interval_start_time=1, interval_end_time=10)
        self.assertEqual([([3], 0), ([5], 1), ([7], 2), ([9], 3)], list(selector.select(data)))

        selector = FrameByIntervalSelector(frames_to_extract=5, interval_start_time=1, interval_end_time=10)
        self.assertEqual([([2], 0), ([4], 1), ([6], 2), ([7], 3), ([9], 4)], list(selector.select(data)))

        selector = FrameByIntervalSelector(frames_to_extract=2, interval_start_time=3, interval_end_time=7)
        self.assertEqual([([4], 0), ([6], 1)], list(selector.select(data)))

    def test_frame_interval_selector_with_framerate(self):
        data = list(range(0, 30 * 5))

        selector = FrameByIntervalSelector(frames_to_extract=1, interval_start_time=1.7, interval_end_time=3.5)
        times = [2.6]
        frames = [([t * 30], i) for i, t in enumerate(times)]
        self.assertEqual(frames, list(selector.select(data, 30)))

        selector = FrameByIntervalSelector(frames_to_extract=3, interval_start_time=1.7, interval_end_time=3.5)
        # times = [2.15, 2.6, 3.05]  # -> [([64.5], 0), ([78.0], 1), ([91.5], 2)]
        frames = [([64], 0), ([78], 1), ([92], 2)]
        self.assertEqual(frames, list(selector.select(data, 30)))

        selector = FrameByIntervalSelector(frames_to_extract=5, interval_start_time=2.1, interval_end_time=4.6)
        # times = [2.516, 2.932, 3.348, 3.76, 4.18]  # -> [([75.48], 0), ([87.96], 1), ([100.44], 2), ([112.8], 3), ([125.39], 4)]
        frames = [([75], 0), ([88], 1), ([101], 2), ([113], 3), ([126], 4)]
        self.assertEqual(frames, list(selector.select(data, 30)))


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod

import numpy as np

import sep.loaders.loader


class Saver(ABC):
    def __init__(self):
        self.annotator = None
        pass

    def close(self):
        pass

    @abstractmethod
    def set_output(self, output_root, loader: sep.loaders.loader.Loader):
        pass

    @abstractmethod
    def save_result(self, name_or_num, result):
        pass

    @abstractmethod
    def save_tag(self, name_or_num, input_tag, result_tag):
        pass



from sep._commons import visuals


def overlay_visual_inspection(image, ground_truth, segmentation, pred_alpha=0.75):
    overlay = visuals.overlay_as_rgb(image, segmentation > ground_truth,
                                     (ground_truth == 1) & (segmentation == 1), ground_truth > segmentation,
                                     pred_alpha=pred_alpha)
    legend = [("True positive", (0, 1, 0)), ("False positive", (1, 0, 0)), ("False negative", (0, 0, 1))]
    return overlay, legend



import itertools
import numpy as np
import numpy.testing as nptest

import sep._commons.movies
from tests.testbase import TestBase


class TestVideoReader(TestBase):
    def test_reader(self):
        video_path = self.root_test_dir("input/reptiles/Dragon - 32109.mp4")
        with sep._commons.movies.StreamReader(video_path) as video:
            self.assertEqual(len(video), 313)
            last_frame = video[len(video) - 1]

            shape = []
            frame = None
            for image in video:
                frame = image
                shape.append(image.shape[0])
            self.assertEqual(1, len(np.unique(shape)))
            nptest.assert_equal(last_frame, frame)

    def test_reader_framerate(self):
        video_path = self.root_test_dir("input/reptiles/Dragon - 32109.mp4")
        with sep._commons.movies.StreamReader(video_path) as video:
            self.assertEqual(len(video), 313)
            self.assertEqual(len(video), len(video.pos_samples(video.frame_rate)))

            self.assertEqual(157, len(video.pos_samples(video.frame_rate / 2.0)))
            self.assertEqual(157, len(list(video.read_samples(video.frame_rate / 2.0))))

    def test_process_as_file(self):
        video_path = self.root_test_dir("input/reptiles/Dragon - 32109.mp4")
        processed_path = self.add_temp("my_movie.mp4")

        thresholded = []
        with sep._commons.movies.StreamReader(video_path) as video:
            with sep._commons.movies.VideoWriter(processed_path, related_reader=video) as writer:
                for image in itertools.islice(video, 0, 5):
                    thresholded.append(image.max(axis=-1) > 128)
                    writer.add(thresholded[-1], accept_bool=True)  # as bool

        with sep._commons.movies.StreamReader(processed_path) as video:
            third_processed = video[2]
            # Beware that this may not be precise dure to compression.
            nptest.assert_equal(thresholded[2], third_processed[..., 0] > 128)

    def test_writer_data_types(self):
        written_path = self.add_temp("my_movie.mp4")

        with sep._commons.movies.VideoWriter(written_path, (100, 100), fps=30) as writer:
            # RGB image
            data_rgb = self.random_rgb((100, 100))
            writer.add(data_rgb)

            # single channel uint8
            data_single = self.random_rgb((100, 100))[..., 0]
            writer.add(data_single)

            # float array
            data_float = np.random.random((100, 100))
            writer.add(data_float, accept_float=True)

            # bool array
            data_bool = np.random.random((100, 100)) > 0.5
            writer.add(data_bool, accept_bool=True)

        with sep._commons.movies.StreamReader(written_path) as video:
            self.assertEqual(4, len(video))

    def test_writer_invalid_size(self):
        written_path = self.add_temp("my_movie.mp4")

        with self.assertRaises(ValueError):
            with sep._commons.movies.VideoWriter(written_path, (100, 100), fps=30) as writer:
                data_rgb = self.random_rgb((60, 60))
                writer.add(data_rgb)



import numpy as np
import numpy.testing as nptest
import unittest

from sep.assessors.regions import EntireRegion, EdgesRegion, DetailsRegion
from tests.testbase import TestBase


class TestRegions(TestBase):
    def test_entire(self):
        entire_region = EntireRegion()
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        self.assertEqual("Entire image", entire_region.name)
        nptest.assert_equal(blob_1, entire_region.regionize(blob_2, blob_1))

    def test_edges_smoke(self):
        edges_region_int = EdgesRegion(2)
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        some_region = edges_region_int.regionize(blob_2, blob_1)

        edges_region_float = EdgesRegion(0.2)
        some_region = edges_region_float.regionize(blob_2, blob_1)

    def test_edges_big(self):
        blob_1 = np.zeros((1080, 1920))
        blob_2 = np.zeros((1080, 1920))

        # TODO test it better: dtypes and resize interpolation
        edges_region_float = EdgesRegion(0.02, downsample_x=640)
        some_region = edges_region_float.regionize(blob_2, blob_1)

        blob_rand_bin_1 = np.random.random((1080, 1920)) > 0.5
        blob_rand_bin_2 = np.random.random((1080, 1920)) > 0.5

        some_region = edges_region_float.regionize(blob_2, blob_1)
        print(some_region.dtype)

    def test_details_smoke(self):
        details_region_int = DetailsRegion(2)
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        some_region = details_region_int.regionize(blob_2, blob_1)

        details_region_float = DetailsRegion(0.2)
        some_region = details_region_float.regionize(blob_2, blob_1)

    def test_expressions(self):
        edges_region_int = EdgesRegion(1)
        blob_1 = np.zeros((10, 10))
        blob_1[0:5, 0:10] = 1

        blob_2 = np.zeros((10, 10))
        blob_2[4:6, 0:5] = 1

        non_edge = ~edges_region_int
        regionized_region_edge = edges_region_int.regionize(blob_2, blob_1)
        regionized_region_non_edge = non_edge.regionize(blob_2, blob_1)
        nptest.assert_equal(regionized_region_edge, regionized_region_non_edge == 0)

        region_edge = edges_region_int.extract_region(blob_2)
        region_non_edge = non_edge.extract_region(blob_2)
        nptest.assert_equal(region_edge, region_non_edge == 0)

        bigger_edges_region_int = EdgesRegion(2)
        or_region = non_edge | bigger_edges_region_int
        and_region = non_edge & bigger_edges_region_int
        region_or = or_region.extract_region(blob_2)
        region_and = and_region.extract_region(blob_2)


if __name__ == '__main__':
    unittest.main()



import os

import collections
import fire
from pathlib import Path

from sep.loaders import FilesLoader


def validate(*split_listings: str, use_file_path=False):
    data_splits = {}
    for listing_path in split_listings:
        listing_name = Path(listing_path).name
        print(f"Listing: {listing_name}...")
        loader = FilesLoader.from_listing("", filepath=listing_path, validate_list=False)
        print(f"... consist of {len(loader)} samples.")
        data_splits[listing_name] = loader

    # Validate duplicates:
    for split in data_splits:
        if use_file_path:
            split_files = data_splits[split].list_images_paths()
        else:
            split_files = [os.path.basename(img) for img in data_splits[split].list_images()]  # TODO ensure that those are names

        duplicated_files = [item for item, count in collections.Counter(split_files).items() if count > 1]
        if duplicated_files:
            print(f"There are duplications in listing {split}:")
            print(f"There are {len(duplicated_files)} duplicates, for example:")
            print(", ".join(list(duplicated_files)[:10]))

    # Validate separation:
    for split1 in data_splits:
        for split2 in data_splits:
            if split1 < split2:  # ensure only once
                if use_file_path:
                    split1_files = data_splits[split1].list_images_paths()
                    split2_files = data_splits[split2].list_images_paths()
                else:
                    split1_files = data_splits[split1].list_images()
                    split2_files = data_splits[split2].list_images()

                overlap = set(split1_files) & set(split2_files)
                if overlap:
                    print(f"There is overlap between listings {split1} and {split2}:")
                    print(f"{len(overlap)} files are overlaping, for example:")
                    print(", ".join(list(overlap)[:10]))


if __name__ == '__main__':
    fire.Fire()



import fire
import napari

import sep.loaders
import sep.assessors.regions
from sep._commons.gui import Inspector, add_review_option, ReviewEnum


def inspect_loader(data_loader):
    with napari.gui_qt():
        inspector = Inspector(data_loader)
        inspector.create_viewer()
        # TODO show and edit tags?
        # TODO show statistics?


def inspect_regions(data_loader: sep.loaders.FilesLoader):
    inspector = Inspector(data_loader)
    with napari.gui_qt():
        inspector.create_viewer()
        regions = sep.assessors.regions.set_standard[1:]
        for region in regions:
            inspector.add_extra_label(str(region), region.extract_region)


def filter_loader_manual(data_loader: sep.loaders.FilesLoader, save_tags=False):
    """
    Start dataset filtering procedure using napari gui of the provided data_loader.
    Mark the unwanted samples as rejected. The data_loader will be reduce not to include those.
    Args:
        data_loader: original data_loader that will be filtered
    """
    print(f"Inspected loaded has {len(data_loader)} samples.")
    inspector = Inspector(data_loader)
    with napari.gui_qt():
        inspector.create_viewer()
        add_review_option(inspector)
    new_tags = inspector.viewer_state['temporary_tags']

    # TODO handle some sort of cancel?
    assert not save_tags, "Not yet implemented"

    all_samples = enumerate(data_loader.list_images())
    rejected_samples = []
    non_rejected_samples = []
    for num, name in all_samples:
        if num in new_tags and new_tags[num].get('review', ReviewEnum.No) == ReviewEnum.Rejected:
            rejected_samples.append(name)
        else:
            non_rejected_samples.append(name)

    print(f"Manual inspection rejected {len(rejected_samples)} samples.")
    if len(rejected_samples):
        print(f"List: {', '.join(rejected_samples)}")
    data_loader.filter_files(non_rejected_samples)
    assert len(data_loader) == len(non_rejected_samples)
    print(f"Now loader has {len(data_loader)} samples.")


def regions_for_dataset(data_root, listing_path):
    loader = sep.loaders.ImagesLoader.from_listing(data_root, listing_path)
    inspect_regions(loader)


def filter_dataset_listing(data_root, listing_path, output_path, verbose=1,
                           add_tag_path=True, ensure_samples_exist=True, preserve_order=True):
    if verbose:
        print(f"Filtering listing file {listing_path}.")
        print(f"The data root for dataset is {data_root}.")
        print(f"The new filtered listing (without rejected samples) will be saved to {output_path}.")

    loader = sep.loaders.ImagesLoader.from_listing(data_root, listing_path,
                                                   verbose=1, validate_list=ensure_samples_exist, preserve_order=preserve_order)
    filter_loader_manual(loader)
    loader.save(output_path, add_tag_path)


if __name__ == '__main__':
    fire.Fire({
        "filter_dataset_listing": filter_dataset_listing,
        "regions_for_dataset": regions_for_dataset
    })



import os
from tqdm import tqdm

from sep.loaders.loader import Loader
from sep.producers.producer import Producer
from sep.savers.saver import Saver


def process(data_loader: Loader, producer: Producer, result_saver: Saver,
            output_root,
            verbose=1):
    if verbose:
        print(f"Processing by {producer} on data from '{data_loader}'.")
        print(f"Results will be saved at {output_root} using {result_saver}.")
        print(f"There are {len(data_loader)} images to process.")

    result_saver.set_output(output_root, data_loader)
    os.makedirs(output_root, exist_ok=True)

    for i in tqdm(range(len(data_loader)), producer.name):
        image = data_loader.load_image(i)
        tag = data_loader.load_tag(i)

        segment, segment_tag = producer.calculate(image, tag)
        result_saver.save_result(i, segment)
        result_saver.save_tag(i, input_tag=tag, result_tag=segment_tag)
    result_saver.close()

    return producer



import numbers
import traceback

import numpy as np
import pafy
import typing as t

import sep.loaders
from sep.loaders.movies import FrameSelector
from sep._commons.movies import StreamReader
from sep._commons.utils import *
from sep.loaders.loader import Loader


class YoutubeLoader(Loader):
    """
    This one loads frames from youtube video and tags them so that can be used in processing or in dataset extraction.
    """
    def __init__(self, video_quality, framerate=None, verbose=0):
        super().__init__()
        self.verbose = verbose
        self.video_quality = video_quality
        self.youtube_urls = []
        self.direct_urls = {}

        self.selector = None
        self.framerate = framerate
        self.video_image_reader: t.Optional[StreamReader] = None

        self.input_paths = {}
        self.annotation_paths = {}
        self.json_tags = {}

        self.input_order = sorted(self.input_paths.keys())

    @classmethod
    def from_urls(cls, urls, video_quality, framerate=None, clips_len=1, clips_skip=0, verbose=0):
        self = cls(video_quality=video_quality, framerate=framerate, verbose=verbose)

        self.selector = sep.loaders.movies.FrameByGroupSelector(clips_len, clips_skip)
        for url in urls:
            self.add_url(url, self.selector, {})

        self.input_order = sorted(self.input_paths.keys())
        return self

    def add_url(self, url, selector: FrameSelector, user_info: dict = None):
        if url not in self.youtube_urls:
            self.youtube_urls.append(url)

        youtube_video = pafy.new(url)
        user_info = user_info or dict()
        streams = [v for v in youtube_video.videostreams if v.notes == self.video_quality and v.extension == 'mp4']
        if not streams:
            print(f"{url} has no video_quality: {self.video_quality} version. Skipping...")
            return False

        info = {'title': youtube_video.title, 'author': youtube_video.author, 'id': youtube_video.videoid}
        movie_path = streams[0].url
        movie_id = youtube_video.videoid
        movie_tag = info
        self.direct_urls[movie_id] = movie_path

        # Process input data movie.
        movie_frames = sep.loaders.MoviesLoader.load_movie_images(movie_path, self.framerate, selector=selector)
        for frame_id, tag in zip(movie_frames['images'], movie_frames['tags']):
            frame_path = f"{movie_path}{frame_id}"
            frame_id = f"{movie_id}{frame_id}"
            tag['id'] = frame_id
            update_with_suffix(tag, movie_tag, prefix=sep.loaders.MoviesLoader.MOVIE_TAG_PREFIX)
            update_with_suffix(tag, user_info, prefix=sep.loaders.MoviesLoader.USER_TAG_PREFIX)

            self.input_paths[frame_id] = frame_path
            self.json_tags[frame_id] = tag

        self.input_order = sorted(self.input_paths.keys())
        return True

    def list_movies(self):
        return self.direct_urls.keys()

    def list_movies_paths(self):
        return self.direct_urls.values()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.close()

    def close(self):
        if self.video_image_reader:
            self.video_image_reader.close()
            self.video_image_reader = None

    def list_images_paths(self):
        return [self.input_paths[p] for p in self.input_order]

    def list_images(self):
        return list(self.input_order)

    def __get_frame_path(self, path_set, name_or_num):
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        if isinstance(name_or_num, str):
            return path_set.get(name_or_num, None)
        else:
            raise NotImplemented(type(name_or_num))

    def load_image(self, name_or_num) -> np.ndarray:
        path_to_frame = self.__get_frame_path(self.input_paths, name_or_num)
        if path_to_frame is None:
            raise Exception(f"{name_or_num} does not exist in the loader.")
        path_to_movie, frame_nr = path_to_frame.rsplit("_", maxsplit=1)
        self.video_image_reader = sep.loaders.MoviesLoader.prepare_reader(self.video_image_reader, path_to_movie)
        return self.video_image_reader[int(frame_nr)]

    def load_tag(self, name_or_num):
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        return self.json_tags.get(name_or_num, None)

    def load_annotation(self, name_or_num) -> t.Optional[pathlib.Path]:
        # I doubt that there will ever be annotation video on youtube :)
        return None

    def get_relative_path(self, name_or_num):
        if isinstance(name_or_num, numbers.Integral):
            name_or_num = self.input_order[name_or_num]
        youtube_id = self.load_tag(name_or_num)['movie_id']
        return os.path.join(youtube_id, name_or_num)

    def __str__(self):
        return f"YoutubeLoader for: {self.youtube_urls}"



import itertools
import numpy as np
import os
import unittest

import sep.loaders
from sep.loaders import YoutubeLoader
from tests.testbase import TestBase


@unittest.skip
class TestYoutubeLoader(TestBase):
    def test_loading(self):
        with YoutubeLoader.from_urls(['https://www.youtube.com/watch?v=_cLFseR-S50'], '720p',
                                     framerate=5, clips_len=5, clips_skip=10) as youtube_loader:
            self.assertEqual(1, len(youtube_loader.list_movies()))
            self.assertEqual(1, len(youtube_loader.list_movies_paths()))

            self.assertEqual('_cLFseR-S50_00000', youtube_loader.list_images()[0])
            self.assertEqual(447, len(youtube_loader.list_images_paths()))

            tag_1 = youtube_loader.load_tag('_cLFseR-S50_00000')
            annotation_1 = youtube_loader.load_annotation('_cLFseR-S50_00000')
            image_1 = youtube_loader.load_image('_cLFseR-S50_00000')
            self.assertArray(image_1, 3, np.uint8)
            self.assertIsNone(annotation_1)
            self.assertSubset(tag_1, {'id': '_cLFseR-S50_00000', 'pos': 0, 'pos_clip': 0, 'clip_nr': 0, 'timestamp': 0.0})
            self.assertSubset(tag_1, {'movie_id': '_cLFseR-S50', 'movie_title': 'Slipknot - Snuff (Violet Orlandi cover)',
                                      'movie_author': 'Violet Orlandi'})

    def test_iterating(self):
        with YoutubeLoader.from_urls(['https://www.youtube.com/watch?v=_cLFseR-S50'], '720p',
                                     framerate=5, clips_len=1, clips_skip=10) as youtube_loader:
            self.assertEqual(122, len(youtube_loader.list_images_paths()))

            for sample in itertools.islice(youtube_loader, 0, 5):
                tag = sample['tag']
                annotation = sample['annotation']
                image = sample['image']
                self.assertArray(image, 3, np.uint8)
                self.assertIsNone(annotation)
                self.assertSubset(tag, {'id': '_cLFseR-S50_00000', 'pos': 0, 'pos_clip': 0, 'clip_nr': 0, 'timestamp': 0.0})
                self.assertSubset(tag, {'movie_id': '_cLFseR-S50', 'movie_title': 'Slipknot - Snuff (Violet Orlandi cover)',
                                        'movie_author': 'Violet Orlandi'})
                return

    def test_sequential_urls(self):
        with YoutubeLoader.from_urls([], '720p', framerate=5, clips_len=10, clips_skip=1) as youtube_loader:
            self.assertEqual(0, len(youtube_loader.list_images_paths()))
            youtube_loader.add_url('https://www.youtube.com/watch?v=_cLFseR-S50',
                                   sep.loaders.FrameByGroupSelector(clips_len=1, clips_skip=10))
            self.assertEqual(122, len(youtube_loader.list_images_paths()))

            sample = youtube_loader[0]
            tag = sample['tag']
            annotation = sample['annotation']
            image = sample['image']
            self.assertArray(image, 3, np.uint8)
            self.assertIsNone(annotation)
            self.assertSubset(tag, {'id': '_cLFseR-S50_00000', 'pos': 0, 'pos_clip': 0, 'clip_nr': 0, 'timestamp': 0.0})
            self.assertSubset(tag, {'movie_id': '_cLFseR-S50', 'movie_title': 'Slipknot - Snuff (Violet Orlandi cover)',
                                    'movie_author': 'Violet Orlandi'})
        with YoutubeLoader('720p', framerate=5) as youtube_loader:
            self.assertEqual(0, len(youtube_loader.list_images_paths()))
            youtube_loader.add_url('https://www.youtube.com/watch?v=_cLFseR-S50',
                                   sep.loaders.FrameByGroupSelector(clips_len=1, clips_skip=10))
            self.assertEqual(122, len(youtube_loader.list_images_paths()))

    def test_relative(self):
        with YoutubeLoader.from_urls(['https://www.youtube.com/watch?v=_cLFseR-S50'], '720p',
                                     framerate=5, clips_len=1, clips_skip=10) as youtube_loader:
            data_names = youtube_loader.list_images()
            self.assertEqual(122, len(data_names))
            self.assertEqual("_cLFseR-S50_00000", data_names[0])
            self.assertEqual(os.path.join("_cLFseR-S50", "_cLFseR-S50_00000"), youtube_loader.get_relative_path(0))
            self.assertEqual(os.path.join("_cLFseR-S50", "_cLFseR-S50_00000"),
                             youtube_loader.get_relative_path("_cLFseR-S50_00000"))

    def test_get_interval_group(self):
        with YoutubeLoader('360p') as youtube_loader:
            selector = sep.loaders.FrameByIntervalSelector(frames_to_extract=10, interval_start_time=1.0, interval_end_time=11.0)
            youtube_loader.add_url('https://www.youtube.com/watch?v=_cLFseR-S50', selector=selector,
                                   user_info={"sex": "woman", "hair": "dark"})
            data_names = youtube_loader.list_images()
            self.assertEqual(10, len(data_names))


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod

import numpy as np


class Metric(ABC):
    """
    This represents a single metric that is calculate for a given pair of labels.
    """

    def __init__(self, name):
        self.name = name

    @abstractmethod
    def calculate(self, segmentation: np.ndarray, ground_truth: np.ndarray) -> float:
        pass

    def show(self, segmentation: np.ndarray, ground_truth: np.ndarray) -> float:
        metric_value = self.calculate(segmentation, ground_truth)
        print(f"{self}: {metric_value:.4f}")

    def __str__(self):
        return self.name


class IouMetric(Metric):
    def __init__(self):
        super().__init__("iou")

    def calculate(self, segmentation, ground_truth) -> float:
        if segmentation.max() > 1:
            segmentation = segmentation > 0
        if ground_truth.max() > 1:
            ground_truth = ground_truth > 0

        return np.sum(np.logical_and(segmentation, ground_truth)) / \
               np.sum(np.logical_or(segmentation, ground_truth)) + 0.00000001


class RMSEMetric(Metric):
    def __init__(self):
        super().__init__("rmse")

    def calculate(self, segmentation, ground_truth) -> float:
        segmentation = segmentation.astype(np.float32)
        ground_truth = ground_truth.astype(np.float32)
        return np.sqrt(np.mean((segmentation - ground_truth) ** 2))



from glob import glob

import itertools
import numbers
import typing as t

from sep._commons.utils import *
from sep.loaders.loader import Loader


class FilesLoader(Loader):
    """
    Look through entire file structure in the data_root path and collect all the images.
    """

    def __init__(self, data_root, verbose=0):
        super().__init__()
        self.verbose = verbose
        self.data_root = data_root
        self.input_paths = {}
        self.annotation_paths = {}
        self.json_tags = {}
        self.input_order = []
        self.extended_ids = False
        self.annotation_for_image_finder = None

    @classmethod
    def from_tree(cls, data_root, input_extensions=None,
                  annotation_suffix="_gt", annotation_extension=".png",
                  annotation_checker: t.Callable[[pathlib.Path], bool] = None,
                  annotation_for_image_finder: t.Callable[[pathlib.Path], str] = None,
                  verbose=0):
        """
        Initialize loader that uses pairs of files as samples.
        First it finds the input images:
        - those in data_root
        - with input_extensions
        - not ending in annotation_suffix

        Then for each input image it looks for the corresponding annotation and tags file.
        By default it looks in the same folder as image:
        - data_001.png - data_001_gt.png - data_001.json

        But this can be customized by providing a annotation_for_image_finder function.
        Args:
            data_root: root folder where the files can be found, if using
            input_extensions: extensions of the input image file (default: [".tif", ".png", ".jpg"])
            annotation_suffix: suffix of the annotation files, if None it is not used in filtering
            annotation_extension: extension of the annotation files, used only when annotation_for_image_finder is None
            annotation_checker: custom function that determines if the given path represents annotations
                if not None then it overrides check based on annotation_suffix
            annotation_for_image_finder: custom function that determines path of the corresponding annotation,
                overrides annotation_extension
            verbose: if non-zero additional summaries are printed
        """
        self = cls(data_root, verbose=verbose)
        self.preserve_order = False
        self.input_extensions = input_extensions or [".tif", ".png", ".jpg"]
        self.annotation_checker = annotation_checker
        if self.annotation_checker is None:
            self.annotation_checker = lambda p: annotation_suffix is not None and p.stem.endswith(annotation_suffix)

        self.annotation_for_image_finder = annotation_for_image_finder
        if self.annotation_for_image_finder is None:
            assert_arg(annotation_extension is not None,
                       "annotation_extension has to be not None if annotation_for_image_finder is None.")
            self.annotation_for_image_finder = lambda p: p.with_name(p.stem + annotation_suffix + annotation_extension)

        all_files = [pathlib.Path(p) for p in sorted(glob(os.path.join(data_root, "**", "*.*"), recursive=True))]

        input_images_paths = [f for f in all_files
                              if f.suffix.lower() in self.input_extensions and not self.annotation_checker(f)]
        annotation_paths = []
        json_tags = []
        for input_path in input_images_paths:
            annotation_paths.append(self.annotation_for_image_finder(input_path))
            json_path = self.suggest_json_path(input_path)
            json_tags.append(json_path)

        self.set_files(input_rel_paths=input_images_paths, annotation_rel_paths=annotation_paths, tag_rel_paths=json_tags)
        return self

    @classmethod
    def from_listing(cls, data_root, filepath, verbose=0, validate_list=False, preserve_order=True):
        """
        Initialize loader that uses pairs of files as samples.
        It uses provided listing file to find those files starting from data_root.

        Args:
            data_root: root folder where the files can be found, if using
            filepath: path to the listing file.
                The listing file consists of lines per each sample:
                    - path to image file
                    - optional path to annotation file
                    - optional path to tag file
                Example: "humans/human_1.tif, humans/human_1_gt.png, humans/human_1.json"
            verbose: if non-zero additional summaries are printed
            validate_list: if True then each of the files in the listing has to exist. Otherwise only input image is checked
            preserve_order: if True then original listing order will be preserved
        """
        self = cls(data_root, verbose=verbose)
        self.listing_filepath = filepath
        self.preserve_order = preserve_order

        with open(filepath, "r") as listing_file:
            samples = [l.strip() for l in listing_file.readlines() if l.strip()]

        input_rel_paths = []
        annotation_rel_paths = []
        tag_rel_paths = []
        for line_sample in samples:
            sample = [t.strip() for t in line_sample.split(",")]
            # make sure that it handles windows slashes on unix
            input_rel_paths.append(ensure_posix_path_str(sample[0]))
            annotation_rel_paths.append(ensure_posix_path_str(None if len(sample) <= 1 else sample[1]))
            tag_rel_paths.append(ensure_posix_path_str(None if len(sample) <= 2 else sample[2]))

        self.set_files(input_rel_paths=input_rel_paths, annotation_rel_paths=annotation_rel_paths, tag_rel_paths=tag_rel_paths,
                       validate_list=validate_list)
        return self

    def set_files(self, *, input_rel_paths, annotation_rel_paths, tag_rel_paths, validate_list=False):
        assert_arg(None not in input_rel_paths, "input_paths has None values")
        assert_arg(len(input_rel_paths) == len(annotation_rel_paths) == len(tag_rel_paths),
                   "All three list should have the same length.")
        self.input_paths = {}
        self.annotation_paths = {}
        self.json_tags = {}

        for input_path, annotation_path, tag_path in itertools.zip_longest(input_rel_paths, annotation_rel_paths, tag_rel_paths):
            input_path = self.data_root / pathlib.Path(input_path)
            sample_id = self.path_to_id(input_path)
            # check for duplicates
            if sample_id in self.input_paths:
                if not self.extended_ids:  # if there is still chance to use extended and get unique ids
                    self.extended_ids = True
                    print("The filenames are colliding. Trying again with extended ids option.")
                    self.set_files(input_rel_paths=input_rel_paths, annotation_rel_paths=annotation_rel_paths,
                                   tag_rel_paths=tag_rel_paths, validate_list=validate_list)
                    return
                else:
                    raise Exception("Failed to determine unique extended ids for each of the files.")

            self.input_paths[sample_id] = input_path
            assert_value(not validate_list or os.path.isfile(input_path), f"Specified image does not exist: f{input_path}")

            if annotation_path is not None:
                annotation_path = self.data_root / pathlib.Path(annotation_path)
                if os.path.isfile(annotation_path):
                    self.annotation_paths[sample_id] = annotation_path
                else:
                    assert_value(not validate_list, f"Specified annotation file does not exist {annotation_path}.")

            if tag_path is not None:
                tag_path = self.data_root / pathlib.Path(tag_path)
                if os.path.isfile(tag_path):
                    self.json_tags[sample_id] = tag_path
                else:
                    assert_value(not validate_list, f"Specified tag file does not exist {tag_path}.")
            # TODO this should save those missing paths so that it can be filled out in Annotator or Tagger

        self.input_order = list(self.input_paths.keys()) if self.preserve_order else sorted(self.input_paths.keys())
        if self.verbose:
            self.show_summary()

    def suggest_json_path(self, input_path):
        # Assume standard json path next to the input file.
        return pathlib.Path(input_path).with_suffix(".json")

    def show_summary(self):
        print(self)
        print(f"Found {len(self.input_paths)} images.")
        print(f"Found {len(self.annotation_paths)} annotations.")
        print(f"Found {len(self.json_tags)} tags.")

    def filter_files(self, names_or_nums):
        new_input_order = []
        for name_or_num in names_or_nums:
            name = self.get_name(name_or_num)
            new_input_order.append(name)
            input_path = self.input_paths.get(name, None)
            assert_arg(input_path is not None, f"{name} does not exist in FileLoader.")
        self.input_order = new_input_order
        self.input_paths = subset(self.input_paths, new_input_order)
        self.annotation_paths = subset(self.annotation_paths, new_input_order)
        self.json_tags = subset(self.json_tags, new_input_order)

    def get_relative_paths(self, name_or_num):
        input_rel_path = self.__get_file_path(self.input_paths, name_or_num, relative=True)
        json_rel_path = self.__get_file_path(self.json_tags, name_or_num, relative=True)
        annotation_rel_path = self.__get_file_path(self.annotation_paths, name_or_num, relative=True)
        return {"image": input_rel_path,
                "tag": json_rel_path,
                "annotation": annotation_rel_path}

    def save(self, listing_path, add_tag_path=True):
        data_lines = []
        for name_or_num in self.list_images():
            paths = self.get_relative_paths(name_or_num)
            line = [paths['image'], paths['annotation']]
            if add_tag_path:
                line.append(paths['tag'])
            line = [p or "" for p in line]
            data_lines.append(", ".join(line) + "\n")

        with open(listing_path, "w") as listing_file:
            listing_file.writelines(data_lines)

    def get_name(self, name_or_num):
        return self.input_order[name_or_num] if isinstance(name_or_num, numbers.Integral) else name_or_num

    def path_to_id(self, path: pathlib.Path):
        # TODO this still may not be unique, we may use ids from tags instead
        if self.extended_ids:
            return path.parent.name + "/" + path.stem
        return path.stem

    def list_images(self):
        return list(self.input_order)

    def list_images_paths(self):
        return [self.input_paths[p] for p in self.input_order]

    def __get_file_path(self, path_set, name_or_num, relative=False):
        name = self.get_name(name_or_num)
        if isinstance(name, str):
            file_path = path_set.get(name, None)
            if relative and file_path is not None:
                file_path = os.path.relpath(file_path, self.data_root)
            return file_path
        else:
            raise NotImplemented(type(name))

    def load_image(self, name_or_num) -> pathlib.Path:
        path_to_file = self.__get_file_path(self.input_paths, name_or_num)
        assert_value(path_to_file is not None, f"Image {name_or_num} does not exist in the loader.")
        return pathlib.Path(path_to_file)

    def save_tag(self, name_or_num, new_tag):
        assert "id" in new_tag

        name = self.get_name(name_or_num)
        path_to_file = self.__get_file_path(self.json_tags, name)
        if path_to_file is None:
            input_path = self.__get_file_path(self.input_paths, name)
            path_to_file = self.suggest_json_path(input_path)
            self.json_tags[name] = path_to_file

        save_json(self.json_tags[name], new_tag)

    def load_tag(self, name_or_num):
        path_to_file = self.__get_file_path(self.json_tags, name_or_num)
        if path_to_file is None:
            return {"id": name_or_num}
        else:
            tag = load_json(path_to_file)
            assert "id" in tag
            return tag

    def get_relative_path(self, name_or_num):
        return self.__get_file_path(self.input_paths, name_or_num, relative=True)

    def load_annotation(self, name_or_num) -> t.Optional[pathlib.Path]:
        path_to_file = self.__get_file_path(self.annotation_paths, name_or_num)
        if path_to_file is None:
            return None
        return pathlib.Path(path_to_file)

    def __str__(self):
        return f"FileLoader for: {self.data_root}"



import imageio
import numpy as np
import typing as t

import sep._commons.imgutil as imgutil
from sep.loaders.files import FilesLoader, SepException


class ImagesLoader(FilesLoader):
    """
    Look through entire file structure in the data_root path and collect all the images.
    It loads input and annotations as np.ndarray.
    """

    def load_image(self, name_or_num) -> np.ndarray:
        path_to_file = super().load_image(name_or_num)
        return imageio.imread(path_to_file)

    def load_annotation(self, name_or_num) -> t.Optional[np.ndarray]:
        path_to_file = super().load_annotation(name_or_num)
        if path_to_file is None:
            return None
        annotation_data = imageio.imread(path_to_file)
        # TODO maybe let user know that it was 3d
        annotation_data = imgutil.make_2d(annotation_data)
        # TODO ensure that size matches to the load image - maybe checking on demand.
        self.validate_annotation(annotation_data)
        return annotation_data

    def save_annotation(self, name_or_num, new_annotation: np.ndarray, keep_history=False, new_annotation_path=None):
        assert not keep_history, "Keeping annotation changes history not yet implemented for ImagesLoader."
        # Check consistency with image.
        image = self.load_image(name_or_num)
        assert imgutil.get_2d_size(image) == imgutil.get_2d_size(new_annotation)
        self.validate_annotation(new_annotation)

        # Optionally update annotation path
        assert new_annotation_path is None, "Explicit new annotation path setting not yet implemented."

        # Save new annotation into the given path.
        path_to_file = super().load_annotation(name_or_num)
        if path_to_file is None:
            # Try to create new annotation file.
            input_path = super().load_image(name_or_num)
            if self.annotation_for_image_finder is None:
                raise SepCannotDetermineAnnotationPathError()

            name = self.get_name(name_or_num)
            annotation_path = self.annotation_for_image_finder(input_path)
            self.annotation_paths[name] = annotation_path
            assert super().load_annotation(name_or_num) == annotation_path
            path_to_file = annotation_path
        imageio.imwrite(str(path_to_file), new_annotation)

    def __str__(self):
        return f"ImageLoader for: {self.data_root}"


class SepCannotDetermineAnnotationPathError(SepException):
    def __init__(self, msg=None):
        if msg is None:
            msg = "There is no way to determine the annotation path:\n" \
                  "\t- set 'annotation_for_image_finder' function\n" \
                  "\t- use loader created from tree\n" \
                  "\t- specify it explicitely\n"
        super(SepCannotDetermineAnnotationPathError, self).__init__(msg)



import numpy as np
import pandas as pd
from matplotlib import cm
from matplotlib import pyplot as plt

from sep._commons.utils import *


class Plotter:
    """
    This class is responsible for plotting and managing tabular results e.g. coming from metricer.
    """

    def __init__(self, fig_size=None):
        self.figure_size = fig_size
        self.plotter_default = {}
        if self.figure_size is not None:
            self.plotter_default['figsize'] = self.figure_size

    def plot_series(self, series: pd.DataFrame, name: str, x_column, y_column, colour,
                    point_size=None, figsize=None, ax=None, kind='scatter', **plot_dict):
        plot_dict = plot_dict or {}
        add_defaults(plot_dict, self.plotter_default)
        specific_params = {}
        if point_size is not None:
            specific_params['s'] = point_size
        if figsize is not None:
            specific_params['figsize'] = figsize
        plot_dict.update(specific_params)

        ax = series.plot(kind=kind, x=x_column, y=y_column, label=name, ax=ax, color=colour,
                         grid=True, **plot_dict)
        return ax

    def plot(self, full_data: pd.DataFrame, groups_column: str, x_column, y_column,
             x_axis_label=None, as_lines=False, **plot_dict):
        plot_dict = plot_dict or {}
        defaults = {'style': '.-', 'ylabel': y_column}
        add_defaults(plot_dict, defaults)

        if as_lines:
            plot_kind = 'line'
            plot_dict = defaults
        else:
            plot_kind = 'scatter'
        cmap = cm.get_cmap('tab10')  # I like it.
        ax = None

        # Check if x_column can be float if not make it labels as plot as ordered.
        unique_xs = None
        x_axis_label = x_axis_label or x_column
        try:
            np.asanyarray(full_data[x_column], float)
        except ValueError:
            unique_xs = full_data[x_column].unique()
            enumerated_xs = [(x, i) for i, x in enumerate(unique_xs)]
            di = dict(enumerated_xs)
            plot_dict['xlabel'] = x_column

            new_x_column = 'order_' + x_column
            full_data[new_x_column] = full_data[x_column].replace(di)
            x_column = new_x_column

        groups = full_data.groupby(groups_column)

        for i, (name, group) in enumerate(groups, 1):
            ax = self.plot_series(group, name, x_column, y_column, colour=cmap([i]), ax=ax, kind=plot_kind, **plot_dict)

        # Pick sensible xtics
        if unique_xs is not None:
            ax.set_xlabel(x_axis_label)
            ax.set_xlim((0, len(unique_xs) - 1))

            ax.set_xticks(list(range(0, len(unique_xs))))
            ax.set_xticklabels(list(unique_xs))
        return ax

    def summary_box(self, full_data: pd.DataFrame, groups_column, y_column, **box_plot_params):
        ax = full_data.boxplot(by=groups_column, column=y_column, **box_plot_params)
        plt.suptitle('')
        return ax



from abc import abstractmethod, ABC
import typing as t

import numpy as np

import skimage.morphology
import skimage.transform

from sep._commons.utils import *


class Region(ABC):
    """
    This class generate the transformations of the segmentation and ground truth so that they can be evaluated
    in the same manner as the entire image. E.g. this can be used to generate metrics on only edges of the ground
    truth mask.
    """

    def __init__(self, name):
        self.name = name
        self.viable_thresh = 0.005

    def regionize(self, ground_truth: np.ndarray, mask: np.ndarray) -> np.ndarray:
        # TODO rethink mask 0-1 vs 0-255 or it may not be a mask?
        relevant_area = self.extract_region(ground_truth)
        return mask.astype(bool) & relevant_area

    @abstractmethod
    def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
        pass

    def area_fractions(self, ground_truth: np.ndarray) -> t.Tuple[float, float]:
        """
        Calculate the region coverage fraction.
        Returns:
        (fraction of the input image in regression, fraction of the ground truth in region against entire input_image)
        """
        region_array = self.extract_region(ground_truth)
        region_fractions = region_array.sum() / ground_truth.size
        annotation_in_region_fractions = ((region_array > 0) * (ground_truth > 0)).sum() / ground_truth.size
        return region_fractions, annotation_in_region_fractions

    def is_viable_for(self, ground_truth: np.ndarray) -> bool:
        reg_frac, _ = self.area_fractions(ground_truth)
        return reg_frac > self.viable_thresh

    def __invert__(self):
        return RegionExpr('~', self)

    def __and__(self, other):
        return RegionExpr('~', self, other)

    def __or__(self, other):
        return RegionExpr('|', self, other)

    def __str__(self):
        return self.name


class RegionExpr(Region):
    def __init__(self, operator, *regions, name=None):
        name = name or f"Expr[{operator}]({regions})"
        super().__init__(name)
        self.operator = operator
        self.regions = regions

    def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
        if self.operator == '|':
            return self.regions[0].extract_region(ground_truth) | self.regions[1].extract_region(ground_truth)
        elif self.operator == '&':
            return self.regions[0].extract_region(ground_truth) & self.regions[1].extract_region(ground_truth)
        elif self.operator == '~':
            return ~self.regions[0].extract_region(ground_truth)
        else:
            assert_arg(False, self.operator)


class EntireRegion(Region):
    def __init__(self):
        super().__init__("Entire image")

    def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
        return np.ones_like(ground_truth, dtype=bool)


class EdgesRegion(Region):
    def __init__(self, edge_size, name="Edges", downsample_x=None):
        """
        Region consisting of the edge of the ground truth.
        Args:
            edge_size: if int it is pixel size, if float it is the fraction of the mean of image dimension
        """
        super().__init__(name)
        self.edge_size = edge_size
        self.downsample_x = downsample_x

    def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
        if self.downsample_x:
            original_shape = ground_truth.shape
            scale = self.downsample_x / ground_truth.shape[1]
            ground_truth = skimage.transform.rescale(ground_truth, scale)

        if isinstance(self.edge_size, float):
            mean_size = (ground_truth.shape[0] + ground_truth.shape[1]) / 2
            selem = skimage.morphology.disk(mean_size * self.edge_size)
        else:
            selem = skimage.morphology.disk(self.edge_size)
        dilated = skimage.morphology.binary_dilation(ground_truth, selem)
        eroded = skimage.morphology.binary_erosion(ground_truth, selem)
        final_region = dilated > eroded
        if self.downsample_x:
            final_region = skimage.transform.resize(final_region, original_shape)
            # Ensure region is binary.
            final_region = final_region > 0.5
        return final_region


class DetailsRegion(Region):
    def __init__(self, edge_size, name="Details"):
        """
        Region consisting of the small objects of the ground truth.
        Args:
            edge_size: if int it is pixel size, if float it is the fraction of the mean of image dimension
        """
        super().__init__(name)
        self.edge_size = edge_size

    def extract_region(self, ground_truth: np.ndarray) -> np.ndarray:
        if isinstance(self.edge_size, float):
            mean_size = (ground_truth.shape[0] + ground_truth.shape[1]) / 2
            selem = skimage.morphology.disk(mean_size * self.edge_size)
        else:
            selem = skimage.morphology.disk(self.edge_size)
        opened = skimage.morphology.binary_opening(ground_truth, selem)
        return (ground_truth > 0) > opened


set_standard = [
    EntireRegion(),
    RegionExpr('~', EdgesRegion(2), name="No edges pixels"),  # Edge pixels are disregarded.
    EdgesRegion(0.02, name="Mask precision", downsample_x=640),  # Difference near the edge.
    RegionExpr('~', EdgesRegion(0.02, downsample_x=640), name="Mask robust"),  # Robustness.
    DetailsRegion(0.05, name="Mask details"),  # Details (hand, arms, hair).
]



import os

import setuptools

# TODO maybe use setuptools_scm to get that and add version.py with current version
version = os.environ.get("RELEASE_VERSION", "0.0.0")

setuptools.setup(
    author="Filip Mroz",
    author_email="fafafft@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering"
    ],
    cmdclass={},
    include_package_data=True,
    keywords=["evaluation", "semantic", "segmentation", "benchmark"],
    license="BSD",
    long_description="",
    name="semevalplatform",
    description="Semantic Evaluation Platform",
    packages=setuptools.find_packages(exclude=[
        "tests", "examples"
    ]),
    install_requires=[
        "numpy",
        "fire",
        "imageio",
        "pathlib",
        "attrdict",
        "tqdm",
        "pandas",
        "matplotlib",
        "notebook",
        "jupyter_contrib_nbextensions",
        "scikit-image",
        "opencv-python",
        "pafy",
        "youtube_dl"
    ],
    url="https://github.com/Fafa87/SEP",
    version=version
)



import imageio
import json
import numpy as np
import os
import pathlib

from sep._commons.utils import *
import sep.loaders.images
from sep.savers.saver import Saver


class ImagesSaver(Saver):
    """
    Save results for images from ImagesLoader in the same hierarchy as the loader.
    """

    def __init__(self, output_root='.', loader: sep.loaders.images.ImagesLoader = None, image_format=".png",
                 verbose=0, ignore_dtype=False):
        super().__init__()
        self.ignore_dtype = ignore_dtype
        self.image_format = image_format
        self.loader = loader
        self.verbose = verbose
        self.output_root = pathlib.Path(output_root)

    def set_output(self, output_root, loader: sep.loaders.images.ImagesLoader):
        self.loader = loader
        self.output_root = pathlib.Path(output_root)

    def save_result(self, name_or_num, result: np.ndarray, ignore_dtype=None):
        assert self.loader, "Loader has to be provided."
        ignore_dtype = self.ignore_dtype if ignore_dtype is None else ignore_dtype
        if ignore_dtype:
            assert result.dtype == np.uint8, \
                "Floats would have to be converted to uint8. If desired add override type param"

        relative_img_path = pathlib.Path(self.loader.get_relative_path(name_or_num))
        result_path = (self.output_root / relative_img_path).with_suffix(self.image_format)
        os.makedirs(result_path.parent, exist_ok=True)
        imageio.imwrite(result_path, result)

    def save_tag(self, name_or_num, input_tag, result_tag):
        assert self.loader, "Loader has to be provided."
        relative_img_path = self.loader.get_relative_path(name_or_num)
        tag_path = (self.output_root / relative_img_path).with_suffix(".json")
        os.makedirs(tag_path.parent, exist_ok=True)
        save_json(tag_path, input_tag, result_tag)



import traceback

import collections
import numpy as np
import os
import pathlib
import sys

import sep.loaders
import sep.loaders.images
from sep._commons.movies import VideoWriter
from sep._commons.utils import *
from sep.savers.saver import Saver


class MoviesSaver(Saver):
    """
    Save results to video files from MoviesLoader in the same hierarchy as the loader.
    It only supports sequential processing in the same order as required MoviesLoader.

    If you want lossless video file you need to use a lossless codecs (eg. FFMPEG FFV1, Huffman HFYU, Lagarith LAGS).
    """

    def __init__(self, output_root='.', loader: sep.loaders.MoviesLoader = None, movie_format=".mp4", encoding='mp4v',
                 verbose=0, ignore_dtype=False):
        super().__init__()
        self.encoding = encoding
        self.ignore_dtype = ignore_dtype
        self.movie_format = movie_format
        self.loader = loader
        self.verbose = verbose
        self.output_root = pathlib.Path(output_root)
        self.video_writer = None
        self.video_writer_order = collections.defaultdict(int)

    def set_output(self, output_root, loader: sep.loaders.MoviesLoader):
        self.loader = loader
        self.output_root = pathlib.Path(output_root)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
        self.close()

    def close(self):
        if self.video_writer:
            self.video_writer.close()
            self.video_writer = None

    def prepare_writer(self, video_writer, path_to_movie):
        if video_writer is not None and video_writer.output_path != str(path_to_movie):
            self.video_writer_order[path_to_movie] = sys.maxsize
            video_writer.close()
            video_writer = None
        if video_writer is None:
            assert self.loader.video_image_reader is not None, \
                "The loader has not started loading data so it is not possible to use it to configure writer."
            video_writer = VideoWriter(path_to_movie, encoding=self.encoding,
                                       fps=self.loader.framerate, output_shape=self.loader.video_image_reader.shape)
            video_writer.__enter__()
        return video_writer

    def save_result(self, name_or_num, result: np.ndarray, ignore_dtype=None):
        assert self.loader, "Loader has to be provided."

        # save into the file that you should save
        ignore_dtype = self.ignore_dtype if ignore_dtype is None else ignore_dtype
        if ignore_dtype:
            assert result.dtype == np.uint8, \
                "Floats would have to be converted to uint8. If desired add override type param"

        # get movie and frame to add
        relative_frame_path = pathlib.Path(self.loader.get_relative_path(name_or_num))
        frame_id = relative_frame_path.stem
        movie_id, frame_nr = self.loader.split_frame_path(frame_id)
        output_movie_path = (self.output_root / movie_id).with_suffix(self.movie_format)

        if movie_id not in self.video_writer_order:
            self.video_writer_order[movie_id] = frame_nr
        else:
            assert self.video_writer_order[movie_id] <= frame_nr, "The images and movies has to be saved sequentially."
            self.video_writer_order[movie_id] = frame_nr

        os.makedirs(output_movie_path.parent, exist_ok=True)
        self.video_writer = self.prepare_writer(self.video_writer, output_movie_path)
        self.video_writer.add(result, True, True)

    def save_tag(self, name_or_num, input_tag, result_tag):
        assert self.loader, "Loader has to be provided."

        # save only movie level information
        relative_frame_path = pathlib.Path(self.loader.get_relative_path(name_or_num))
        frame_id = relative_frame_path.stem
        movie_id, frame_nr = self.loader.split_frame_path(frame_id)
        output_tag_path = (self.output_root / movie_id).with_suffix(".json")

        # TODO check if it is not chainging
        input_tag = get_with_prefix(input_tag, sep.loaders.MoviesLoader.MOVIE_TAG_PREFIX)
        save_json(output_tag_path, input_tag, result_tag)



import numpy as np
import numpy.testing as nptest

import sep.inspect.regions
import tests.testbase


class TestRegions(tests.testbase.TestBase):
    def test_overlay_single(self):
        random_soup = np.random.random((10, 10, 3)) / 2.0
        empty_mask = np.zeros(random_soup.shape[:-1], dtype=np.uint8)

        overlay = sep.inspect.regions.overlay_region(random_soup, empty_mask)
        are_greys = overlay[...,0] == overlay[...,1]
        self.assertEqual(np.count_nonzero(are_greys), overlay.size / 3)

        full_mask = np.ones(random_soup.shape[:-1], dtype=np.uint8)
        overlay = sep.inspect.regions.overlay_region(random_soup, full_mask)
        are_greys = overlay[..., 0] == overlay[..., 1]
        self.assertEqual(np.count_nonzero(are_greys), 0)

    def test_overlay_multiple(self):
        whiteness = np.ones((10, 10, 3)) * 70
        half_mask = np.zeros(whiteness.shape[:-1], dtype=np.uint8)
        half_mask[0:5] = 1
        second_mask = np.zeros(whiteness.shape[:-1], dtype=np.uint8)
        second_mask[4:10] = 1

        overlay, legend = sep.inspect.regions.overlay_regions(whiteness, [half_mask, second_mask], regions_names=None)
        pass

